/**
 * @file
 * @brief Main Window class implementation
 */
#include <iostream>
#include <fstream>
#include <sstream>
#include <cstring>

#pragma GCC diagnostic ignored "-Wold-style-cast"
#pragma GCC diagnostic ignored "-Wzero-as-null-pointer-constant"
#pragma GCC diagnostic ignored "-Weffc++"
#pragma GCC diagnostic ignored "-Wsign-conversion"
#pragma GCC diagnostic ignored "-Wconversion"
#pragma GCC diagnostic ignored "-Wcast-qual"
#pragma GCC diagnostic ignored "-Wfloat-conversion"
#include <TApplication.h>
#include <TGLabel.h>
#include <TGButton.h>
#include <TGTab.h>
#include <TGComboBox.h>
#include <TGProgressBar.h>
#include <TGSpeedo.h>
#include <TGCommandPlugin.h>
#include <TGHtml.h>
#include <TGSplitter.h>
#include <TGToolBar.h>
#include <TGFileDialog.h>
#include <TGNumberEntry.h>
//#include <TGTextEntry.h>
//#include <TGTextView.h>
//#include <TGClient.h>
#include <TImage.h>
#include <TColor.h>
#include <TEnv.h>
#include <TROOT.h>
#include <TMath.h>
#include <TStopwatch.h>
#include <TSystem.h>
//#include <TThread.h>
#include <Getline.h>
#include <TRootEmbeddedCanvas.h>
#include <TCanvas.h>
//#include <TRint.h>
//#include <TSystem.h>
#include <TPRegexp.h>
#include "MainWindow.hpp"
#include "ThSFMC01.hpp"
#include "PulseSurfer.hpp"
#pragma GCC diagnostic warning "-Wold-style-cast"
#pragma GCC diagnostic warning "-Wzero-as-null-pointer-constant"
#pragma GCC diagnostic warning "-Weffc++"
#pragma GCC diagnostic warning "-Wsign-conversion"
#pragma GCC diagnostic warning "-Wconversion"
#pragma GCC diagnostic warning "-Wcast-qual"
#pragma GCC diagnostic warning "-Wfloat-conversion"

//#include "RegisterDict.hpp"
//#include "BoardRegister.hpp"
#include "utils.hpp"
#ifdef ZDTUSB
#include "Keysight.hpp"
#endif

//const uint32_t kSensorUpdateTime = 1000;    ///< Sensor update time in ms
const uint32_t kSensorAverage = 100;        ///< Buffer size for moving average calculation on sensors
const uint32_t kBufferAverage = 8;          ///< Buffer size for moving average calculation on chunk fill
//const uint32_t kRegisterUpdateTime = 1000;  ///< Register update time ms
const uint32_t kClockUpdateTime = 1000;     ///< Clock update time ms
#ifdef ZDTUSB
const uint32_t kScreenshotUpdateTime = 1500;///< Sensor update time in ms
#endif
const uint32_t kQuitTimeout = 100;          ///< Quit timeout in ms
const uint32_t kStartTimeout = 100;         ///< Start timeout in ms

MainWindow::MainWindow(const TGWindow * const p, const UInt_t w, const UInt_t h, PulseSurfer* const ps, const uint32_t vendor, const uint32_t prod, const bool quit) :
    TGMainFrame(p, w, h),
    fSurfer(ps),
    fSurferChannel(-1),
    fThCont(new ThSFMC01(0)),
    fQuit(quit)
#ifdef ZDTUSB
    , fKeysight(vendor == 0 ? nullptr : new Keysight(vendor, prod))
#endif
{
#ifndef ZDTUSB//Silence warnings
    (void)(vendor);
    (void)(prod);
#endif
    SetCleanup(kDeepCleanup);//Set automatic cleanup

    Connect("CloseWindow()", "MainWindow", this, "DoClose()");
    DontCallClose();

    /*if(fThCont->IsOpen())
    {
        fThCont->ResetMaster();
    }*/

    TGLayoutHints* const topleft = new TGLayoutHints(kLHintsTop | kLHintsLeft, 5, 5, 5, 5);
    TGLayoutHints* const expandX = new TGLayoutHints(kLHintsExpandX);
    TGLayoutHints* const lCols = expandX;
    TGLayoutHints* const lBut = new TGLayoutHints(kLHintsExpandX, 5, 5, 5, 5);
    TGLayoutHints* const lChkbox = new TGLayoutHints(kLHintsNormal, 0, 0, 7, 3);
    TGLayoutHints* const lCombox = new TGLayoutHints(kLHintsNormal, 0, 0, 7, 3);
    TGLayoutHints* const lLab    = new TGLayoutHints(kLHintsLeft | kLHintsExpandX, 0, 0, 7, 3);
    TGLayoutHints* const lSpeedo = new TGLayoutHints(kLHintsExpandX, -11, -11, 0, 0);
    TGLayoutHints* const lSpeedo2 = new TGLayoutHints(kLHintsExpandX, 0, 0, 0, 0);
    TGLayoutHints* const lProg = lLab;
    TGLayoutHints* const lHtml = new TGLayoutHints(kLHintsExpandX | kLHintsExpandY, 5, 5, 2, 2);
    TGLayoutHints* const lSplit = new TGLayoutHints(kLHintsExpandY, 5, 5, 2, 2);
    TGLayoutHints* const lComm = new TGLayoutHints(kLHintsExpandX | kLHintsExpandY, 5, 5, 2, 2);
    TGLayoutHints* const lTool = new TGLayoutHints(kLHintsExpandX | kLHintsTop);
    TGLayoutHints* const lPic = new TGLayoutHints(kLHintsNormal, 5, 5, 2, 2);

    fTab = new TGTab(this, 630, 500);//https://root-forum.cern.ch/t/gui-using-tgtab/21468
    AddFrame(fTab, new TGLayoutHints(kLHintsBottom | kLHintsExpandX | kLHintsExpandY, 2, 2, 5, 1));
    fTab->SetScrollingEnabled();//wait for it to be on ROOT master branch
    fTab->Connect("Selected(Int_t)","MainWindow",this,"HandleTabs(Int_t)");

    TGCompositeFrame* const daqFrame = fTab->AddTab("Data Acquisition");
    TGHorizontalFrame* const daqColumns = new TGHorizontalFrame(daqFrame, 630, 600, kFixedWidth);//600+5*4+5*2
    daqFrame->AddFrame(daqColumns, topleft);

    TGVerticalFrame* const daqCol1 = new TGVerticalFrame(daqColumns, 200, 200, kFixedWidth);
    daqColumns->AddFrame(daqCol1, topleft);

    TGGroupFrame* const sampling = new TGGroupFrame(daqCol1, new TGString("Sampling control"), kVerticalFrame | kRaisedFrame);
    daqCol1->AddFrame(sampling, lCols);

    fRisingEdge = new TGCheckButton(sampling,"Rising Edge");
    fRisingEdge->SetToolTipText("Enable Trigger on Rising Edge");
    sampling->AddFrame(fRisingEdge, lChkbox);
    fFallingEdge = new TGCheckButton(sampling,"Falling Edge");
    fFallingEdge->SetToolTipText("Enable Trigger on Falling Edge");
    sampling->AddFrame(fFallingEdge, lChkbox);
    fExtTrigDelay = new TGCheckButton(sampling,"Ext. Trig. Delay");
    fExtTrigDelay->SetToolTipText("Enable External Trigger Delay");
    sampling->AddFrame(fExtTrigDelay, lChkbox);
    fSignedMode = new TGCheckButton(sampling,"Dec/Avg Signed Mode");
    fSignedMode->SetToolTipText("Enable decimation/average signed mode");
    sampling->AddFrame(fSignedMode, lChkbox);
    fFrequency = new TGComboBox(sampling, "Sampling frequency...");
    fFrequency->EnableTextInput(kFALSE);
    fFrequency->Resize(130,fSignedMode->GetHeight());
    fFrequency->AddEntry("2.0 GSPS",0);
    fFrequency->AddEntry("2.5 GSPS",1);
    sampling->AddFrame(fFrequency, lCombox);
    fDecimation = new TGComboBox(sampling, "Decimation...");
    fDecimation->EnableTextInput(kFALSE);
    fDecimation->Resize(fFrequency->GetWidth(),fSignedMode->GetHeight());
    fDecimation->AddEntry("No decimation",0);
    fDecimation->AddEntry("Decimation = 2",2);
    fDecimation->AddEntry("Decimation = 4",4);
    sampling->AddFrame(fDecimation, lCombox);
    fAveraging = new TGCheckButton(sampling,"Averaging");
    fAveraging->SetToolTipText("Enable waveform averaging");
    sampling->AddFrame(fAveraging, lChkbox);
    fEventHeader = new TGCheckButton(sampling,"Event Header");
    fEventHeader->SetToolTipText("Store header for each trigger (if Continuous Mode disabled)");
    sampling->AddFrame(fEventHeader, lChkbox);
    fContinuous = new TGCheckButton(sampling,"Continuous");
    fContinuous->SetToolTipText("Enable continuous sampling mode");
    sampling->AddFrame(fContinuous, lChkbox);
    fBinary = new TGCheckButton(sampling,"Binary");
    fBinary->SetToolTipText("Enable Binary unsigned ADC data format instead of default signed Two's complement");
    sampling->AddFrame(fBinary, lChkbox);
    fTestMode = new TGComboBox(sampling, "Select test mode...");
    fTestMode->EnableTextInput(kFALSE);
    fTestMode->Resize(fFrequency->GetWidth(),fSignedMode->GetHeight());
    fTestMode->AddEntry("Normal operation",0);
    fTestMode->AddEntry("Midscale short",1);
    fTestMode->AddEntry("Positive full scale",2);
    fTestMode->AddEntry("Negative full scale",3);
    fTestMode->AddEntry("Checker board",4);
    fTestMode->AddEntry("Random long",5);
    fTestMode->AddEntry("Random short",6);
    fTestMode->AddEntry("1/0 word toggle",7);
    fTestMode->AddEntry("User pattern",8);
    fTestMode->AddEntry("Ramp output",15);
    sampling->AddFrame(fTestMode, lCombox);

    fLocked = new TGLabel(sampling,"PLL unlocked");//PLL locked kGreen
    fLocked ->SetTextColor(TColor::Number2Pixel(kRed));
    sampling->AddFrame(fLocked ,lLab);
    fOverrange = new TGLabel(sampling,"Overrange");//Within range kGreen
    fOverrange ->SetTextColor(TColor::Number2Pixel(kRed));
    sampling->AddFrame(fOverrange ,lLab);
    fArmed = new TGLabel(sampling,"Disarmed");//Armed kGreen+1
    fArmed->SetTextColor(TColor::Number2Pixel(kOrange));
    sampling->AddFrame(fArmed,lLab);
    fSamplingBusy = new TGLabel(sampling,"Sampling Idle");//Busy kOrange
    fSamplingBusy->SetTextColor(TColor::Number2Pixel(kGreen+1));
    sampling->AddFrame(fSamplingBusy,lLab);
    fBank1Busy = new TGLabel(sampling,"Bank1 Idle");//Busy kOrange
    fBank1Busy->SetTextColor(TColor::Number2Pixel(kGreen+1));
    sampling->AddFrame(fBank1Busy,lLab);
    fBank2Busy = new TGLabel(sampling,"Bank2 Idle");//Busy kOrange
    fBank2Busy->SetTextColor(TColor::Number2Pixel(kGreen+1));
    sampling->AddFrame(fBank2Busy,lLab);
    UpdateLabels();

    fConfigure = new TGTextButton(daqCol1, "&Configure");
    fConfigure->SetToolTipText("Configure sampling and clock");
    fConfigure->Connect("Clicked()", "MainWindow", this, "DoConfigureSampling()");
    daqCol1->AddFrame(fConfigure, lBut);

    for(auto cb : {fRisingEdge, fFallingEdge, fExtTrigDelay, fAveraging, fSignedMode, fEventHeader, fContinuous, fBinary})
    {
        cb->Connect("Clicked()","TGTextButton", fConfigure, "SetEnabled()");
        cb->Connect("Clicked()","TGTextButton", fConfigure, "ChangeText(=\"&Configure\")");//https://github.com/root-project/root/issues/7195
    }
    for(auto cb : {fFrequency, fDecimation, fTestMode})
    {
        cb->Connect("Changed()","TGTextButton", fConfigure, "SetEnabled()");
        cb->Connect("Changed()","TGTextButton", fConfigure, "ChangeText(=\"&Configure\")");
        //https://github.com/root-project/root/issues/7195
    }

    fExit = new TGTextButton(daqCol1, "&Exit");
    fExit->Connect("Clicked()", "MainWindow", this, "DoClose()");
    //fExit->Connect("Clicked()", "TApplication", gApplication, "Terminate()");
    daqCol1->AddFrame(fExit, lBut);

    TGVerticalFrame *daqCol[2];
    TGGroupFrame* grch[2];
    for(unsigned short ch = 0; ch <= 1; ++ch)
    {
        daqCol[ch] = new TGVerticalFrame(daqColumns, 200, 200, kFixedWidth);
        daqColumns->AddFrame(daqCol[ch], topleft);
        grch[ch] = new TGGroupFrame(daqCol[ch], new TGString((static_cast<TString>("Channel ")+(ch==0?"A":"B")).Data()), kVerticalFrame | kRaisedFrame);
        daqCol[ch]->AddFrame(grch[ch], lCols);

        fStore[ch] = new TGCheckButton(grch[ch],"Store");
        fStore[ch]->SetToolTipText("Store ADC data");
        grch[ch]->AddFrame(fStore[ch], lChkbox);

        fSource[ch] = new TGComboBox(grch[ch], "Source...");
        fSource[ch]->EnableTextInput(kFALSE);
        fSource[ch]->Resize(150,fAveraging->GetHeight());
        fSource[ch]->AddEntry(static_cast<TString>("Source = ch")+(ch==0?"A":"B"),0);
        fSource[ch]->AddEntry(static_cast<TString>("Source = ch")+(ch==0?"A":"B")+" compr.",2);
        fSource[ch]->AddEntry(static_cast<TString>("Source = ch")+(ch==0?"B":"A")+" (test)",3);
        grch[ch]->AddFrame(fSource[ch], lCombox);

        fCompress[ch] = new TGCheckButton(grch[ch],"Compress");
        fCompress[ch]->SetToolTipText("Compress ADC data");
        grch[ch]->AddFrame(fCompress[ch], lChkbox);

        fTestCompress[ch] = new TGCheckButton(grch[ch],"Test Compress");
        fTestCompress[ch]->SetToolTipText("Enable Compression Test");
        grch[ch]->AddFrame(fTestCompress[ch], lChkbox);

        fSaveGPIO[ch] = new TGCheckButton(grch[ch],"Save GPIO");
        fSaveGPIO[ch]->SetToolTipText("Store GPIO status on ADC data");
        grch[ch]->AddFrame(fSaveGPIO[ch], lChkbox);

        fGPIOSelect[ch] = new TGComboBox(grch[ch], "Select GPIO...");
        fGPIOSelect[ch]->EnableTextInput(kFALSE);
        fGPIOSelect[ch]->Resize(fSource[ch]->GetWidth(),fAveraging->GetHeight());
        fGPIOSelect[ch]->AddEntry("GPIO = TS_IN",0);
        fGPIOSelect[ch]->AddEntry("GPIO = TRIG_IN",1);
        fGPIOSelect[ch]->AddEntry("GPIO = USER1_IN",2);
        fGPIOSelect[ch]->AddEntry("GPIO = USER2_IN",3);
        fGPIOSelect[ch]->AddEntry("GPIO = USER3_IN",4);
        //fGPIOSelect[ch]->Select(3+ch,kFALSE);
        grch[ch]->AddFrame(fGPIOSelect[ch], lCombox);

        fStartBlock[ch] = new TGComboBox(grch[ch], "Block start address...");
        fStartBlock[ch]->EnableTextInput(kFALSE);
        fStartBlock[ch]->Resize(fSource[ch]->GetWidth(),fAveraging->GetHeight());
        fStartBlock[ch]->AddEntry("Start: 0x0000000",0);
        fStartBlock[ch]->AddEntry("Start: 0x1000000",1);
        fStartBlock[ch]->Select(ch,kFALSE);
        grch[ch]->AddFrame(fStartBlock[ch], lCombox);

        fChunkSize[ch] = new TGComboBox(grch[ch], "DMA read chunk size...");
        fChunkSize[ch]->EnableTextInput(kFALSE);
        fChunkSize[ch]->Resize(fSource[ch]->GetWidth(),fAveraging->GetHeight());
        fChunkSize[ch]->AddEntry("Chunk = 256 MiB",0);
        fChunkSize[ch]->AddEntry("Chunk = 512 MiB",1);
        grch[ch]->AddFrame(fChunkSize[ch], lCombox);

        fNextFill[ch] = new TGSpeedo(grch[ch],0.,200., "Full", "[%]");
        fNextFill[ch]->SetDisplayText("Read", "Chunks");
        fNextFill[ch]->SetThresholds(0.5,0.75,1.0);
        fNextFill[ch]->EnableThreshold();
        fNextFill[ch]->EnableMeanMark();
        fNextFill[ch]->EnablePeakMark();
        fNextFill[ch]->SetBufferSize(kBufferAverage);
        grch[ch]->AddFrame(fNextFill[ch], lSpeedo);

        //"RAM buffer size..."
        TGHorizontalFrame *br = new TGHorizontalFrame(grch[ch], 100, fStartBlock[ch]->GetHeight());
        grch[ch]->AddFrame(br, new TGLayoutHints(kLHintsExpandX));

        fBufferSize[ch] = new TGNumberEntry(br, 1, 3, ch, TGNumberFormat::EStyle::kNESInteger, TGNumberFormat::EAttribute::kNEANonNegative, TGNumberFormat::ELimit::kNELLimitMinMax, 0, 255);
        fBufferSize[ch]->Resize(100,fAveraging->GetHeight());
        br->AddFrame(fBufferSize[ch], lCombox);

        TGLabel* lr = new TGLabel(br, "GiB RAM");
        br->AddFrame(lr, new TGLayoutHints(kLHintsLeft, 5, 0, 8, 2));

        fBufferFill[ch] = new TGHProgressBar(grch[ch], 100, fStartBlock[ch]->GetHeight());
        fBufferFill[ch]->ShowPosition(kTRUE);
        fBufferFill[ch]->SetBarColor("lightblue");
        grch[ch]->AddFrame(fBufferFill[ch], lProg);

        fTransfSpeed[ch] = new TGHProgressBar(grch[ch], 100, fStartBlock[ch]->GetHeight());
        fTransfSpeed[ch]->SetRange(0,6000);
        fTransfSpeed[ch]->ShowPosition(kTRUE, kFALSE, "%.0f MiB/s");
        fTransfSpeed[ch]->SetBarColor(TColor::Number2Pixel(kOrange));
        grch[ch]->AddFrame(fTransfSpeed[ch], lProg);

        if(ch==0)
        {
            fReserveMem = new TGTextButton(daqCol[ch], "&Reserve Memory");
            fReserveMem->Connect("Clicked()", "MainWindow", this, "DoReserveMemory()");
            fReserveMem->SetToolTipText("Allocate RAM memory buffers");
            daqCol[ch]->AddFrame(fReserveMem, lBut);

            fArming = new TGTextButton(daqCol[ch], "&Arm");
            fArming->Connect("Clicked()", "MainWindow", this, "DoArm()");
            fArming->SetToolTipText("Arm logic");
            daqCol[ch]->AddFrame(fArming, lBut);

            fSwTrig = new TGTextButton(daqCol[ch], "&Software Trigger");
            fSwTrig->SetToolTipText("Start data acquisition via software trigger");
            fSwTrig->Connect("Clicked()", "MainWindow", this, "DoSoftwareTrigger()");
            daqCol[ch]->AddFrame(fSwTrig, lBut);

            fDisarm = new TGTextButton(daqCol[ch], "&Disarm Logic");
            fDisarm->SetToolTipText("Disarm logic");
            fDisarm->Connect("Clicked()", "MainWindow", this, "DoDisarmLogic()");
            daqCol[ch]->AddFrame(fDisarm, lBut);
        }
        else if(ch==1)
        {
            fClock = new TGLabel(daqCol[ch],"00:00.00");//,TGLabel::GetDefaultGC()(),TGLabel::GetDefaultFontStruct(),kChildFrame,TColor::Number2Pixel(kBlack));
            fClock->SetTextFont(gClient->GetFontByName("-adobe-helvetica-bold-r-*-*-22-*-*-*-*-*-iso8859-1"));
            fClock->SetTextColor(TColor::Number2Pixel(kBlue));
            fClock->SetSize({100,100});
            daqCol[ch]->AddFrame(fClock,lBut);
            fClockUpdater.Connect("Timeout()", "MainWindow", this, "UpdateClock()");
            fClockUpdater.Connect("Timeout()", "MainWindow", this, "UpdateLabels()");
        }

        fStore[ch]->Connect("Clicked()","TGTextButton", fReserveMem, "SetEnabled()");
        fStore[ch]->Connect("Clicked()","TGTextButton", fReserveMem, "ChangeText(=\"&Reserve Memory\")");//https://github.com/root-project/root/issues/7195
        fStore[ch]->Connect("Clicked()","TGTextButton", fConfigure, "SetEnabled()");
        fStore[ch]->Connect("Clicked()","TGTextButton", fConfigure, "ChangeText(=\"&Configure\")");//https://github.com/root-project/root/issues/7195
        fBufferSize[ch]->Connect("Modified()","TGTextButton", fReserveMem, "SetEnabled()");
        fBufferSize[ch]->Connect("Modified()","TGTextButton", fReserveMem, "ChangeText(=\"&Reserve Memory\")");
        fChunkSize[ch]->Connect("Changed()","TGTextButton", fReserveMem, "SetEnabled()");
        fChunkSize[ch]->Connect("Changed()","TGTextButton", fReserveMem, "ChangeText(=\"&Reserve Memory\")");
        //https://github.com/root-project/root/issues/7195


        fCompress[ch]->Connect("Clicked()","TGTextButton", fConfigure, "SetEnabled()");
        fCompress[ch]->Connect("Clicked()","TGTextButton", fConfigure, "ChangeText(=\"&Configure\")");//https://github.com/root-project/root/issues/7195
        fTestCompress[ch]->Connect("Clicked()","TGTextButton", fConfigure, "SetEnabled()");
        fTestCompress[ch]->Connect("Clicked()","TGTextButton", fConfigure, "ChangeText(=\"&Configure\")");//https://github.com/root-project/root/issues/7195

        for(auto cb : {fSource[ch], fGPIOSelect[ch], fStartBlock[ch]})
        {
            cb->Connect("Changed()","TGTextButton", fConfigure, "SetEnabled()");
            cb->Connect("Changed()","TGTextButton", fConfigure, "ChangeText(=\"&Configure\")");
            //https://github.com/root-project/root/issues/7195
        }

        fMaxBytes[ch] = 0;
    }

    TGVerticalFrame* daqC = daqCol[1];
    fDAQControls = new TGToolBar(daqC);
    daqC->AddFrame(fDAQControls, lTool);

    ToolBarData_t play;
    play.fPixmap = "bld_redo.xpm";
    play.fTipText = "Start DAQ Thread";
    play.fStayDown = kFALSE;
    play.fId = 0;
    play.fButton = nullptr;
    fDAQControls->AddButton(daqC, &play, 5);
    fDAQControls->GetButton(0)->Connect("Clicked()","MainWindow",this,"DoStartDAQ()");

    ToolBarData_t pause;
    pause.fPixmap = "bld_grab.xpm";
    pause.fTipText = "Pause DAQ Thread";
    pause.fStayDown = kFALSE;
    pause.fId = 1;
    pause.fButton = nullptr;
    fDAQControls->AddButton(daqC, &pause, 5);
    fDAQControls->GetButton(1)->Connect("Clicked()","MainWindow",this,"DoPauseDAQ()");

    ToolBarData_t stop;
    stop.fPixmap = "bld_stop.xpm";
    stop.fTipText = "Stop DAQ Thread";
    stop.fStayDown = kFALSE;
    stop.fId = 2;
    stop.fButton = nullptr;
    fDAQControls->AddButton(daqC, &stop, 5);
    fDAQControls->GetButton(2)->Connect("Clicked()","MainWindow",this,"DoStopDAQ()");

    ToolBarData_t save;
    save.fPixmap = "bld_save.xpm";
    save.fTipText = "Store data to disk";
    save.fStayDown = kFALSE;
    save.fId = 3;
    save.fButton = nullptr;
    fDAQControls->AddButton(daqC, &save, 5);
    fDAQControls->GetButton(3)->SetEnabled(kFALSE);
    fDAQControls->GetButton(3)->Connect("Pressed()","MainWindow",this,"DoSaveData(=false)");

    ToolBarData_t toggle;
    toggle.fPixmap = "bld_replace.xpm";
    toggle.fTipText = "Toggle Pulse Surfer channel";
    toggle.fStayDown = kFALSE;
    toggle.fId = 4;
    toggle.fButton = nullptr;
    fDAQControls->AddButton(daqC, &toggle, 5);
    fDAQControls->GetButton(4)->SetEnabled(kFALSE);
    fDAQControls->GetButton(4)->Connect("Pressed()","MainWindow",this,"ToggleSurfer()");


    // Enable compression when writing to disk
    fWriteCompression = new TGComboBox(daqC, "Write compression...");
    fWriteCompression->EnableTextInput(kFALSE);
    fWriteCompression->Resize(120,fSignedMode->GetHeight());
    fWriteCompression->AddEntry("No compression",0);
    fWriteCompression->AddEntry("GZ",1);
    fWriteCompression->AddEntry("LZ4",2);
    daqC->AddFrame(fWriteCompression, lCombox);

    fOutFile = new TGLabel(daqC, "");
    daqC->AddFrame(fOutFile, new TGLayoutHints(kLHintsExpandX));

    TGCompositeFrame* const senFrame = fTab->AddTab("Sensors");

    TGVerticalFrame* const vSens = new TGVerticalFrame(senFrame, 630, 600);
    senFrame->AddFrame(vSens,topleft);

    TGGroupFrame* const temps = new TGGroupFrame(vSens, new TGString("Temperature monitoring"), kHorizontalFrame | kRaisedFrame);
    vSens->AddFrame(temps, lCols);
    for( uint16_t ch = 0; ch < 3; ++ch )
    {
        fSensors[ch] = new TGSpeedo(temps,0.,100., "Temp", std::to_string(ch).c_str());
        fSensors[ch]->SetDisplayText("Temp", "C");
        fSensors[ch]->SetThresholds(30,50,70);
        fSensors[ch]->EnableThreshold();
        fSensors[ch]->EnableMeanMark();
        fSensors[ch]->EnablePeakMark();
        fSensors[ch]->SetBufferSize(kSensorAverage);
        temps->AddFrame(fSensors[ch], lSpeedo2);
    }
    TGGroupFrame* const rpms = new TGGroupFrame(vSens, new TGString("Ventilator monitoring"), kHorizontalFrame | kRaisedFrame);
    vSens->AddFrame(rpms, lCols);
    for( uint16_t ch = 14; ch < 17; ++ch )
    {
        fSensors[ch] = new TGSpeedo(rpms,0.,20000., "Speed", std::to_string(ch).c_str());
        fSensors[ch]->SetDisplayText("Speed", "rpm");
        fSensors[ch]->SetThresholds(5000,10000,15000);
        fSensors[ch]->EnableThreshold();
        fSensors[ch]->EnableMeanMark();
        fSensors[ch]->EnablePeakMark();
        fSensors[ch]->SetBufferSize(kSensorAverage);
        rpms->AddFrame(fSensors[ch], lSpeedo2);
    }

    //https://root-forum.cern.ch/t/scrollbar-on-tgverticalframe/9788/5
    TGCanvas *fCan = new TGCanvas(vSens, 600, 250, kFixedSize);
    vSens->AddFrame(fCan, new TGLayoutHints(kLHintsExpandY | kLHintsExpandX));
    TGVerticalFrame *fVFrame = new TGVerticalFrame(fCan->GetViewPort(), 10, 10);
    fCan->SetContainer(fVFrame);

    TGGroupFrame* const vlts = new TGGroupFrame(fVFrame, new TGString("Voltage monitoring"), kHorizontalFrame | kRaisedFrame);
    fVFrame->AddFrame(vlts, lCols);
    for( uint16_t ch = 3; ch < 14; ++ch )
    {
        fSensors[ch] = new TGSpeedo(vlts,0.,12000., "Voltage", std::to_string(ch).c_str());
        fSensors[ch]->SetDisplayText("Voltage", "mV");
        fSensors[ch]->SetThresholds(3000,6000,9000);
        fSensors[ch]->EnableThreshold();
        fSensors[ch]->EnableMeanMark();
        fSensors[ch]->EnablePeakMark();
        fSensors[ch]->SetBufferSize(kSensorAverage);
        vlts->AddFrame(fSensors[ch], lSpeedo2);
    }
    fSensorUpdater.Connect("Timeout()", "MainWindow", this, "UpdateSensors()");

    TGCompositeFrame* regFrame;
    TGVerticalFrame* regV;
    TGHtml* html;
    TGHorizontalFrame* butts;
    TGPictureButton* sbut;

    regFrame = fTab->AddTab("SIS1160");
    regV = new TGVerticalFrame(regFrame, 200, 200);
    regFrame->AddFrame(regV, lHtml);
    butts = new TGHorizontalFrame(regV);
    regV->AddFrame(butts, expandX);
    sbut = new TGPictureButton(butts,"bld_redo.xpm");
    sbut->SetToolTipText("Restart Register updater");
    sbut->Connect("Clicked()","MainWindow",this,"HandleTabs(=2)");
    butts->AddFrame(sbut, lPic);
    sbut = new TGPictureButton(butts,"bld_grab.xpm");
    sbut->SetToolTipText("Pause Register updater");
    sbut->Connect("Clicked()","TTimer",&fRegisterUpdater,"Stop()");
    butts->AddFrame(sbut, lPic);
    html = new TGHtml(regV, 600,700);
    regV->AddFrame(html, lHtml);
    fHtmlSis = html;

    // @todo monitor clock frequency? PCIe status, etc

    regFrame = fTab->AddTab("SFMC01");
    regV = new TGVerticalFrame(regFrame, 200, 200);
    regFrame->AddFrame(regV, lHtml);
    butts = new TGHorizontalFrame(regV);
    regV->AddFrame(butts, expandX);
    sbut = new TGPictureButton(butts,"bld_redo.xpm");
    sbut->SetToolTipText("Restart Register updater");
    sbut->Connect("Clicked()","MainWindow",this,"HandleTabs(=3)");
    butts->AddFrame(sbut, lPic);
    sbut = new TGPictureButton(butts,"bld_grab.xpm");
    sbut->SetToolTipText("Pause Register updater");
    sbut->Connect("Clicked()","TTimer",&fRegisterUpdater,"Stop()");
    butts->AddFrame(sbut, lPic);
    html = new TGHtml(regV, 600,700);
    regV->AddFrame(html, lHtml);
    fHtmlSfmc = html;

    regFrame = fTab->AddTab("AD9689");
    regV = new TGVerticalFrame(regFrame, 200, 200);
    regFrame->AddFrame(regV, lHtml);
    butts = new TGHorizontalFrame(regV);
    regV->AddFrame(butts, expandX);
    sbut = new TGPictureButton(butts,"bld_redo.xpm");
    sbut->SetToolTipText("Restart Register updater");
    sbut->Connect("Clicked()","MainWindow",this,"HandleTabs(=4)");
    butts->AddFrame(sbut, lPic);
    sbut = new TGPictureButton(butts,"bld_grab.xpm");
    sbut->SetToolTipText("Pause Register updater");
    sbut->Connect("Clicked()","TTimer",&fRegisterUpdater,"Stop()");
    butts->AddFrame(sbut, lPic);
    html = new TGHtml(regV, 600,700);
    regV->AddFrame(html, lHtml);
    fHtmlAdc = html;

    regFrame = fTab->AddTab("HMC7044");
    regV = new TGVerticalFrame(regFrame, 200, 200);
    regFrame->AddFrame(regV, lHtml);
    butts = new TGHorizontalFrame(regV);
    regV->AddFrame(butts, expandX);
    sbut = new TGPictureButton(butts,"bld_redo.xpm");
    sbut->SetToolTipText("Restart Register updater");
    sbut->Connect("Clicked()","MainWindow",this,"HandleTabs(=5)");
    butts->AddFrame(sbut, lPic);
    sbut = new TGPictureButton(butts,"bld_grab.xpm");
    sbut->SetToolTipText("Pause Register updater");
    sbut->Connect("Clicked()","TTimer",&fRegisterUpdater,"Stop()");
    butts->AddFrame(sbut, lPic);
    html = new TGHtml(regV, 600,700);
    regV->AddFrame(html, lHtml);
    fHtmlHmc = html;

    regFrame = fTab->AddTab("Si570");
    regV = new TGVerticalFrame(regFrame, 200, 200);
    regFrame->AddFrame(regV, lHtml);
    butts = new TGHorizontalFrame(regV);
    regV->AddFrame(butts, expandX);
    sbut = new TGPictureButton(butts,"bld_redo.xpm");
    sbut->SetToolTipText("Restart Register updater");
    sbut->Connect("Clicked()","MainWindow",this,"HandleTabs(=6)");
    butts->AddFrame(sbut, lPic);
    sbut = new TGPictureButton(butts,"bld_grab.xpm");
    sbut->SetToolTipText("Pause Register updater");
    sbut->Connect("Clicked()","TTimer",&fRegisterUpdater,"Stop()");
    butts->AddFrame(sbut, lPic);
    TGHorizontalFrame *fHf = new TGHorizontalFrame(regV, 300, 100);
    regV->AddFrame(fHf, new TGLayoutHints(kLHintsRight | kLHintsExpandX | kLHintsExpandY));
    TGVerticalFrame *fV1 = new TGVerticalFrame(fHf, 300, 100, kFixedWidth);
    TGVerticalFrame *fV2 = new TGVerticalFrame(fHf, 300, 100, kFixedWidth);
    fHf->AddFrame(fV1, new TGLayoutHints(kLHintsLeft | kLHintsExpandY));
    fHf->AddFrame(fV2, new TGLayoutHints(kLHintsRight | kLHintsExpandX | kLHintsExpandY));
    TGVSplitter *splitter = new TGVSplitter(fHf,2,2);
    splitter->SetFrame(fV1, kTRUE);
    fHf->AddFrame(splitter, new TGLayoutHints(kLHintsLeft | kLHintsExpandY));
    fHtmlSi570_1 = new TGHtml(fV1, 600,700);
    fV1->AddFrame(fHtmlSi570_1, lSplit);
    fHtmlSi570_2 = new TGHtml(fV2, 600,700);
    fV2->AddFrame(fHtmlSi570_2, lSplit);

    regFrame = fTab->AddTab("Si5326");
    regV = new TGVerticalFrame(regFrame, 200, 200);
    regFrame->AddFrame(regV, lHtml);
    butts = new TGHorizontalFrame(regV);
    regV->AddFrame(butts, expandX);
    sbut = new TGPictureButton(butts,"bld_redo.xpm");
    sbut->SetToolTipText("Restart Register updater");
    sbut->Connect("Clicked()","MainWindow",this,"HandleTabs(=7)");
    butts->AddFrame(sbut, lPic);
    sbut = new TGPictureButton(butts,"bld_grab.xpm");
    sbut->SetToolTipText("Pause Register updater");
    sbut->Connect("Clicked()","TTimer",&fRegisterUpdater,"Stop()");
    butts->AddFrame(sbut, lPic);
    html = new TGHtml(regV, 600,700);
    regV->AddFrame(html, lHtml);
    fHtmlSi5326 = html;

    fRegisterUpdater.Connect("Timeout()", "MainWindow", this, "UpdateRegisters()");

    TGCompositeFrame* const terFrame = fTab->AddTab("Commands");
    TGCommandPlugin* plug = new TGCommandPlugin(terFrame, 600, 300);
    plug->SetHistAdd();//https://github.com/root-project/root/issues/7180
    Gl_histinit(gEnv->GetValue("Rint.History", gSystem->HomeDirectory()));
    terFrame->AddFrame(plug, lComm);

    #ifdef ZDTUSB
    if(fKeysight && fKeysight->IsOpen())
    {
        TGCompositeFrame* const usbFrame = fTab->AddTab("33611A");
        fScreenshot = new TRootEmbeddedCanvas("Kcanvas",usbFrame, 480, 272, kFixedSize);
        usbFrame->AddFrame(fScreenshot, new TGLayoutHints(kLHintsExpandY | kLHintsExpandX));

        fScreenshot->GetCanvas()->SetLeftMargin(0);
        fScreenshot->GetCanvas()->SetRightMargin(0);
        fScreenshot->GetCanvas()->SetTopMargin(0);
        fScreenshot->GetCanvas()->SetBottomMargin(0);
        //fScreenshot->GetCanvas()->SetSupportGL(kTRUE);
        fScreenshot->SetBit(kCannotPick);
        //~ fScreenshot->GetCanvas()->SetBit(kCannotPick);

        //gStyle->SetOptStat(0);
        //gStyle->SetOptTitle(0);

        fPicture = TImage::Create();

        fScreenshotUpdater.Connect("Timeout()", "MainWindow", this, "UpdateScreenshot()");
        //to-do: make it more fancy with regions: https://root.cern.ch/doc/v614/WorldMap_8C_source.html
    }
    #endif

    //Make the controller class 'accessible' via the CLING interpreter with the global variable fCont
    //gROOT->ProcessLine("#include <ThSFMC01.hpp>");//https://root-forum.cern.ch/t/error-in-tinterpreter-autoload-failure-loading-library/43591
    //https://stackoverflow.com/questions/24494761/c-stringstream-inline
    gROOT->ProcessLine(static_cast<TString>("ThSFMC01* const fThCont = reinterpret_cast<ThSFMC01*>(") + static_cast<std::ostringstream&>(std::ostringstream("") << fThCont).str() + ");");
    #ifdef ZDTUSB
    if(fKeysight && fKeysight->IsOpen())
    {
        gROOT->ProcessLine("#include <../../zdt-daq/usb/Keysight.hpp>");//https://root-forum.cern.ch/t/error-in-tinterpreter-autoload-failure-loading-library/43591
        gROOT->ProcessLine(static_cast<TString>("Keysight* const fKeysight = reinterpret_cast<Keysight*>(") + static_cast<std::ostringstream&>(std::ostringstream("") << fKeysight).str() + ");");
    }
    #endif

    for(uint16_t ch = 0; ch <= 1; ++ch)
    {
        fDMASize[ch] = 0;
        fRAMSize[ch] = 0;
        fBufferRAM[ch] = nullptr;
    }

    fThCont->Connect("ChunkRead(ULong64_t, ULong64_t, ULong64_t, ULong64_t, ULong64_t, ULong64_t)", "MainWindow", this, "UpdateMonitors(ULong64_t, ULong64_t, ULong64_t, ULong64_t, ULong64_t, ULong64_t)");
    if(fQuit)
    {
        fThCont->Connect("Finished()", "MainWindow", this, "SaveAndExit()");
        fQuitter.Connect("Timeout()", "MainWindow", this, "DoClose()");
    }

    SetWindowName("Zero Dead Time DAQ GUI");
    MapSubwindows();
    Resize(GetDefaultSize());
    MapWindow();
}

MainWindow::~MainWindow()
{
    // Destructor.
    Cleanup();
}

void MainWindow::UpdateClock()
{
    //std::cout << "Updating clock" << std::endl;

    {
        R__LOCKGUARD(fGuiMutex) // scoped mutex
        if(fTab->GetCurrent()==0)
        {
            const Long64_t elapsedTime = (fClockUpdater.GetAbsTime() - fStartTime);
            fClock->SetText(format_duration(elapsedTime - fPausedTime).c_str());
        }
    }
}

void MainWindow::UpdateLabels()
{
    bool locked = false;
    bool armed = false;
    bool busy = false;
    bool bank1 = false;
    bool bank2 = false;
    bool open = false;
    bool overrange = true;
    {
        //R__LOCKGUARD(fThCont->fMutex)
        /*if(fThCont->IsOpen())
        {
            open = true;
            locked = fThCont->IsLocked_ADC_PLL() && fThCont->IsLocked_HMC_PLL();
            fThCont->GetAcquisitionControlStatus(armed, busy, bank1, bank2);
            overrange = fThCont->GetOverrange();
        }*/
    }
    if(open)
    {
        R__LOCKGUARD(fGuiMutex)
        const TString newLocked = locked ? "PLL locked" : "PLL unlocked";
        if(fLocked->GetText()->CompareTo(newLocked))
        {
            fLocked->SetTextColor(locked ? TColor::Number2Pixel(kGreen+1) : TColor::Number2Pixel(kRed));
            fLocked->ChangeText(newLocked);
        }
        const TString newOver = overrange ? "Overrange" : "Within range";
        if(fOverrange->GetText()->CompareTo(newOver))
        {
            fOverrange->SetTextColor(overrange ? TColor::Number2Pixel(kRed) : TColor::Number2Pixel(kGreen+1));
            fOverrange->ChangeText(newOver);
        }
        const TString newArmed = armed ? "Armed" : "Disarmed";
        if(fArmed->GetText()->CompareTo(newArmed))
        {
            fArmed->SetTextColor(armed ? TColor::Number2Pixel(kGreen+1) : TColor::Number2Pixel(kOrange));
            fArmed->ChangeText(newArmed);
        }
        if(!armed)
        {
            if(fArming)
            {
                if(!fArming->IsEnabled())
                {
                    fArming->SetEnabled();
                }
                if(fArming->GetText()->CompareTo("Arm"))//when comparing, & is skipped
                {
                    fArming->ChangeText("&Arm");
                }
            }
        }
        const TString newBusy = busy ? "Sampling Busy" : "Sampling idle";
        if(fSamplingBusy->GetText()->CompareTo(newBusy))
        {
            fSamplingBusy->SetTextColor(busy ? TColor::Number2Pixel(kOrange) : TColor::Number2Pixel(kGreen+1));
            fSamplingBusy->ChangeText(newBusy);
        }
        const TString newBank1 = bank1 ? "Bank 1 Busy" : "Bank 1 Idle";
        if(fBank1Busy->GetText()->CompareTo(newBank1))
        {
            fBank1Busy->SetTextColor(bank1 ? TColor::Number2Pixel(kOrange) : TColor::Number2Pixel(kGreen+1));
            fBank1Busy->ChangeText(newBank1);
        }
        const TString newBank2 = bank2 ? "Bank 2 Busy" : "Bank 2 Idle";
        if(fBank2Busy->GetText()->CompareTo(newBank2))
        {
            fBank2Busy->SetTextColor(bank2 ? TColor::Number2Pixel(kOrange) : TColor::Number2Pixel(kGreen+1));
            fBank2Busy->ChangeText(newBank2);
        }
    }
}

void MainWindow::UpdateSensors()
{
    //std::cout << "Updating sensors" << std::endl;
    float values[17];

    /*{
        //R__LOCKGUARD(fThCont->fMutex) // scoped mutex
        for(unsigned short ch = 0; ch <= 16; ++ch)
        {
            values[ch] = fThCont->GetSensorValue(ch);
        }
    }*/

    {
        R__LOCKGUARD(fGuiMutex) // scoped mutex
        for(unsigned short ch = 0; ch <= 16; ++ch)
        {
            float value = std::round(values[ch]);
            if(ch>=3 && ch < 14)
            {
                value *= 1000;
            }
            fSensors[ch]->SetOdoValue(static_cast<int>(value));
            fSensors[ch]->SetScaleValue(value);
        }
    }

}

void MainWindow::UpdateRegisters(Int_t t)
{
    if(t==-1)
    {
        t = fTab->GetCurrent();
    }
    return;
}

void MainWindow::UpdateMonitors(ULong64_t ch, ULong64_t cNum, ULong64_t writePos, ULong64_t dFill, ULong64_t readTime, ULong64_t startTime)
{
    if(cNum > 0)
    {
        std::cout << "Updating monitors ch" << ch << " readChunks: " << cNum << " wr: " << std::hex << writePos << " dF: " << dFill << std::dec << " rT: " << readTime << " sT: " << startTime << std::endl;
    }
    else
    {
        std::cout << "Resetting monitors ch" << ch << std::endl;
    }

    if(ch > 1)
    {
        std::cout << "Wrong ch" << std::endl;
        return;
    }
    R__LOCKGUARD(fGuiMutex)
    if(fRAMSize[ch]!=0)
    {
        fBufferFill[ch]->SetPosition(static_cast<float>(100*writePos/fRAMSize[ch]));
    }
    fNextFill[ch]->SetOdoValue(static_cast<Int_t>(cNum));
    if(fDMASize[ch]!=0)
    {
        fNextFill[ch]->SetScaleValue(static_cast<float>(100*dFill/fDMASize[ch]));
    }
    if(readTime>0)
    {
        fTransfSpeed[ch]->SetPosition(static_cast<float>(1000ULL*fDMASize[ch]/readTime)/1000.f);
    }
    else
    {
        fTransfSpeed[ch]->SetPosition(0);
    }
}

void MainWindow::SaveAndExit()
{
    DoDisarmLogic();
    DoStopDAQ();
    DoSaveData(true);
    fQuitter.Start(kQuitTimeout, kTRUE);//consider replacing with TTimer::SingleShot();
}

void MainWindow::UpdateScreenshot()
{
}

void MainWindow::HandleTabs(Int_t idx)
{
    if(idx == 0)
    {
        if(gSystem->GetListOfTimers()->IndexOf(&fClockUpdater)!=-1)//If it is running
        {
            UpdateClock();
        }
        UpdateLabels();
    }
    else
    {
        fSensorUpdater.Stop();
    }
    if(idx == 1)
    {
        /*if(fThCont->IsOpen())
        {
            UpdateSensors();
            fSensorUpdater.Start(kSensorUpdateTime);
        }*/
    }

    {
        fSensorUpdater.Stop();
    }
    if(idx == 2 || idx == 3 || idx ==4 || idx == 5 || idx == 6 || idx == 7)
    {

        {
            const std::string newText =
            "<!DOCTYPE html>"
            "<html>"
            "<body>"
            "SIS1160 board is not connected."
            "</body>"
            "</html>";
            TGHtml* html = (idx == 2) ? fHtmlSis : (idx == 3) ? fHtmlSfmc : (idx == 4) ? fHtmlAdc : (idx == 5) ? fHtmlHmc : (idx == 6) ? fHtmlSi570_1  : (idx == 7) ? fHtmlSi5326 : nullptr;
            if ((html->GetText() == nullptr && newText.size()>0) || newText != html->GetText())
            {
                html->Clear();
                html->ParseText(const_cast<char*>(newText.c_str()));
                html->Layout();
            }
            if(idx == 6)
            {
                html = fHtmlSi570_2;
                if ((html->GetText() == nullptr && newText.size()>0) || newText != html->GetText())
                {
                    html->Clear();
                    html->ParseText(const_cast<char*>(newText.c_str()));
                    html->Layout();
                }
            }
        }
    }
    else
    {
        fRegisterUpdater.Stop();
    }
    //std::cout << "Handle" << idx << std::endl;
}

void MainWindow::DoConfigureSampling()
{
    uint16_t frequency, decimation, testMode, sourceA, sourceB, gpioA, gpioB;
    bool rising, falling, delay, averaging, signedMode, storeA, storeB, header, continuous, compressA, testCompressA, compressB, testCompressB, saveGPIOa, saveGPIOb;
    bool binary;
    uint32_t startA, startB;
    {
        R__LOCKGUARD(fGuiMutex)
        fConfigure->SetEnabled(kFALSE);
        fConfigure->SetText("Configuring...");
        frequency = fFrequency->GetSelected()==1?2500:2000;
        rising = fRisingEdge->GetState();
        falling = fFallingEdge->GetState();
        delay = fExtTrigDelay->GetState();
        decimation = static_cast<uint16_t>(std::max(0,fDecimation->GetSelected()));
        averaging = fAveraging->GetState();
        signedMode = fSignedMode->GetState();
        storeA = fStore[0]->GetState();
        storeB = fStore[1]->GetState();
        header = fEventHeader->GetState();
        continuous = fContinuous->GetState();
        sourceA = static_cast<uint16_t>(std::max(0,fSource[0]->GetSelected()));
        sourceB = static_cast<uint16_t>(std::max(0,fSource[1]->GetSelected()));
        compressA = fCompress[0]->GetState();
        testCompressA = fTestCompress[0]->GetState();
        compressB = fCompress[1]->GetState();
        testCompressB = fTestCompress[1]->GetState();
        saveGPIOa = fSaveGPIO[0]->GetState();
        saveGPIOb = fSaveGPIO[1]->GetState();
        gpioA = static_cast<uint16_t>(std::max(0,fGPIOSelect[0]->GetSelected()));
        gpioB = static_cast<uint16_t>(std::max(0,fGPIOSelect[1]->GetSelected()));
        binary = fBinary->GetState();
        testMode = static_cast<uint16_t>(std::max(0,fTestMode->GetSelected()));
        startA = static_cast<uint32_t>(std::max(0,fStartBlock[0]->GetSelected())) << 24;
        startB = static_cast<uint32_t>(std::max(0,fStartBlock[1]->GetSelected())) << 24;
    }
    {
        //R__LOCKGUARD(fThCont->fMutex)
        if(true)
        {
            bool res = true;

            R__LOCKGUARD(fGuiMutex)
            if(res)
            {
                fConfigure->SetText("Configured");
                fConfigure->SetEnabled(kFALSE);
            }
            else
            {
                fConfigure->SetText("Configure failed");
                fConfigure->SetEnabled(kFALSE);
            }
        }
    }
    UpdateLabels();
}

void MainWindow::DoArm()
{
    {
        //R__LOCKGUARD(fThCont->fMutex)
        if(true)
        {
            bool res = true;
            R__LOCKGUARD(fGuiMutex)
            if(res)
            {
                fArming->SetText("Armed");
                fArming->SetEnabled(kFALSE);
            }
            else
            {
                fArming->SetText("Arm failed");
            }
        }
        else
        {
            R__LOCKGUARD(fGuiMutex)
            if(fArming)
            {
                fArming->SetText("Device not open");
                fArming->SetEnabled(kFALSE);
            }
        }
    }
    UpdateLabels();
}

void MainWindow::DoSoftwareTrigger()
{

}


void MainWindow::DoKeysightTrigger()
{

}

void MainWindow::DoDisarmLogic()
{
    bool open = false;
    {

    }
    {
        R__LOCKGUARD(fGuiMutex)
        if(open)
        {
            if(fArming)
            {
                fArming->SetText("&Arm");
                fArming->SetEnabled(kTRUE);
            }
        }
        else
        {
            if(fDisarm)
            {
                fDisarm->SetText("Device not open");
                fDisarm->SetEnabled(kFALSE);
            }
        }
    }
    UpdateLabels();
}

void MainWindow::DoReserveMemory()
{
    uint64_t totalMemory = 0;//in bytes
    uint64_t oldRAM[2] = {0, 0};
    {
        R__LOCKGUARD(fGuiMutex)
        if(fSurferChannel != -1)
        {
            fSurfer->Freeze();
            fSurferChannel = -1;
        }
        if((fBufferSize[0]->GetNumber()==0 && fStore[0]->GetState()) || (fBufferSize[1]->GetNumber()==0 && fStore[1]->GetState()))
        {
            for(uint16_t ch = 0; ch <= 1; ++ch)
            {
                fDMASize[ch] = 0;
                if(fBufferRAM[ch]!=nullptr && fBufferSize[ch]->GetNumber()==0 && fStore[ch]->GetState())
                {
                    R__LOCKGUARD(fGuiMutex)
                    fReserveMem->SetText("Deallocating memory...");
                    std::cout << "Deallocating RAM for channel " << ch << std::flush;
                    delete[] fBufferRAM[ch];
                    std::cout << "." << std::endl;
                    fReserveMem->SetText("Select Buffer Size");
                    fReserveMem->SetEnabled(kFALSE);
                }
                fBufferRAM[ch] = nullptr;
            }
            return;
        }
        const uint64_t ramA = fStore[0]->GetState() ? static_cast<uint64_t>(TMath::Nint(fBufferSize[0]->GetNumber()))*1024UL*1024UL*1024UL : 0;//in bytes
        const uint64_t ramB = fStore[1]->GetState() ? static_cast<uint64_t>(TMath::Nint(fBufferSize[0]->GetNumber()))*1024UL*1024UL*1024UL : 0;//in bytes
        oldRAM[0] = fRAMSize[0];
        oldRAM[1] = fRAMSize[1];
        fRAMSize[0] = ramA;
        fRAMSize[1] = ramB;
        totalMemory = (ramA+ramB);//in bytes
    }

    // Get Memory information
    //https://github.com/root-project/root/blob/4d2a8650ed0f47fbffed33488426aaa6b6dfc73b/core/unix/src/TUnixSystem.cxx#L5153
    //https://github.com/root-project/root/issues/7196
    uint64_t avail = 0;//in bytes
    if (totalMemory != 0)
    {
        FILE *f = fopen("/proc/meminfo", "r");
        if(f)
        {
            TString s;
            while (s.Gets(f)) {
                  if (s.BeginsWith("MemAvailable")) {
                     TPRegexp("^.+: *([^ ]+).*").Substitute(s, "$1");//it is reported as kiB
                     avail = static_cast<uint64_t>(s.Atoi())*1024UL;//bytes
                  }
            }
        }
    }

    const uint64_t oldMemory = oldRAM[0] + oldRAM[1];
    const uint64_t extraMemory = static_cast<uint64_t>(std::max(static_cast<int64_t>(totalMemory) - static_cast<int64_t>(oldMemory), 0L));
    if(totalMemory != 0)
    {
        if(extraMemory < avail)
        {
            R__LOCKGUARD(fGuiMutex)
            fReserveMem->SetEnabled(kFALSE);
            fReserveMem->SetText("Allocating " + std::to_string(totalMemory/1024/1024/1024) + " GiB memory...");
        }
        else
        {
            std::cout << "Insufficient RAM memory available (GiB) " << avail/1024/1024/1024 << " vs " << extraMemory/1024/1024/1024 << std::endl;
            totalMemory = 0;
            fRAMSize[0] = 0;
            fRAMSize[1] = 0;
        }
    }

    for(uint16_t ch = 0; ch <= 1; ++ch)
    {
        if(fRAMSize[ch] != 0)
        {
            R__LOCKGUARD(fGuiMutex)
            fDMASize[ch] = (static_cast<uint32_t>(1) << (std::max(0,fChunkSize[ch]->GetSelected())+28));
        }
        else
        {
            fDMASize[ch] = 0;
            if(fBufferRAM[ch]!=nullptr)
            {
                R__LOCKGUARD(fGuiMutex)
                fReserveMem->SetText("Deallocating memory...");
                std::cout << "Deallocating RAM for channel " << ch << std::flush;
                delete[] fBufferRAM[ch];
                std::cout << "." << std::endl;
            }
            fBufferRAM[ch] = nullptr;
        }
    }

    //Allocate memory
    for(uint16_t ch = 0; ch < 2; ++ch)
    {
        if(fRAMSize[ch] != 0 && fRAMSize[ch]!=oldRAM[ch])
        {
            if(fBufferRAM[ch]!=nullptr)
            {
                std::cout << "Deallocating RAM for channel " << ch << std::flush;
                delete[] fBufferRAM[ch];
                std::cout << "." << std::endl;
            }
            std::cout << "Allocating " << fRAMSize[ch]/1024/1024/1024 << " GiB RAM for channel " << ch << std::flush;
            TStopwatch ramTime;
            ramTime.Start();
            fBufferRAM[ch] = new uint32_t[fRAMSize[ch]/sizeof(uint32_t)]();//initialize values to zero
            //fBufferRAM[ch] = new uint32_t[fRAMSize[ch]/sizeof(uint32_t)];//this is faster, but moves work to later?
            ramTime.Stop();
            std::cout << " in " << ramTime.RealTime() << " seconds (" << (static_cast<Double_t>(fRAMSize[ch]/1024/1024)/ramTime.RealTime()) << " MiB/s)." << std::endl;
        }
    }

    {
        R__LOCKGUARD(fGuiMutex)
        if(totalMemory!=0)
        {
            fReserveMem->SetText(std::to_string(totalMemory/1024/1024/1024)+" GiB allocated");
            fDAQControls->GetButton(4)->SetEnabled(kTRUE);
        }
        else
        {
            fReserveMem->SetText("No memory allocated");
            fReserveMem->SetEnabled(kFALSE);
        }
    }
}

void MainWindow::DoStartDAQ()
{
    fClockUpdater.Start(kClockUpdateTime);
    if(fPauseTime != TTime(0))
    {
        fPausedTime = TTime(fPausedTime) + (fClockUpdater.GetAbsTime() - fPauseTime);
    }
    else
    {
        fStartTime = fClockUpdater.GetAbsTime();
        fPausedTime = 0;
        UpdateClock();
        //Reset counters
        for(uint16_t ch=0; ch<=1; ++ch)
        {
            fNextFill[ch]->SetBufferSize(0);//clear
            fNextFill[ch]->SetMeanValue(0);
            fNextFill[ch]->SetScaleValue(0);
            fNextFill[ch]->ResetPeakVal();
            fNextFill[ch]->SetBufferSize(kBufferAverage);
            UpdateMonitors(ch,0,0,0,0,0);
        }
    }
    //Flush data to static TThread-accesible memory. Not really needed, could be non-static members
    for(uint16_t ch=0; ch<=1; ++ch)
    {
        fThCont->kDMASize[ch] = fDMASize[ch];
        fThCont->kRAMSize[ch] = fRAMSize[ch];
        fThCont->kBufferRAM[ch] = fBufferRAM[ch];
    }

    if(true)
    {
        fThCont->Start();
    }
    fDAQControls->GetButton(3)->SetEnabled(kFALSE);
}

void MainWindow::DoPauseDAQ()
{
    if(gSystem->GetListOfTimers()->IndexOf(&fClockUpdater)==-1)//DAQ is not running
    {
        return;
    }
    if(true)
    {
        fThCont->Pause();
    }
    {
        fClockUpdater.Stop();
        fPauseTime = fClockUpdater.GetAbsTime();
    }
}

void MainWindow::DoStopDAQ()
{
    if(gSystem->GetListOfTimers()->IndexOf(&fClockUpdater)==-1 && fPauseTime != TTime(0))//DAQ is not running or paused
    {
        return;
    }
    {
        fClockUpdater.Stop();
        fPauseTime = 0;
        fDAQControls->GetButton(3)->SetEnabled(kTRUE);
    }
    if(true)
    {
        fThCont->Stop();
    }
}

void MainWindow::DoSaveData(const bool batch)
{

    bool compress = false;
    const TString csel = fWriteCompression->GetSelectedEntry()->GetTitle();
    if(csel != "No compression" && csel != "")
    {
        compress = true;
        if(csel != "GZ")
        {
            std::cout << "Compression " << csel << " has not been implemented yet" << std::endl;
            return;
        }
    }

    TString outputDirectory = fOutFile->GetText()->GetString();
    if(!batch)
    {
        TGFileInfo fi;
        fi.SetFilename(outputDirectory);
        fi.SetIniDir(outputDirectory);
        fi.SetMultipleSelection(kFALSE);
        new TGFileDialog(gClient->GetRoot(), this, kDSave, &fi);
        outputDirectory = fi.fIniDir;
    }
    if(outputDirectory=="")//Including user pressed cancel
    {
        std::cout << "No directory was provided." << std::endl;
        return;
    }

    void* checkIn=gSystem->OpenDirectory(outputDirectory);
    if(!checkIn)
    {
        std::cout << "Error in DoSaveData: '" << outputDirectory << "' can not be accessed. Check typing errors or permissions. Folder has to exist." << std::endl;
        gSystem->FreeDirectory(checkIn);
        fOutFile->ChangeText("");
        return;
    }
    gSystem->FreeDirectory(checkIn);//We have to free it for avoiding memory leak.

    fOutFile->ChangeText(outputDirectory);

    for(uint16_t ch = 0; ch <=1 ; ++ch)
    {
        if(fRAMSize[ch]==0)
        {
            continue;
        }
        //@todo TTimeStamp automatically as folder name?
        const TString fileName = outputDirectory + "/ch" + (ch == 0 ? "A" : "B") + (compress ? ".bin.gz" : ".bin");
        if(gSystem->AccessPathName(fileName, kReadPermission)==kFALSE)//bizarre return convention
        {
            std::cout << "Warning, file " << fileName << " already exists, overwriting" << std::endl;
        }
        uint64_t byteLength = fDMASize[ch]*static_cast<uint64_t>(fNextFill[ch]->GetOdoVal());//@todo handle remaining data in half-buffer, if implemented in Stop function.
        if(byteLength == 0)
        {
            continue;
        }
        if(fMaxBytes[ch] != 0 && fMaxBytes[ch] < byteLength)
        {
            byteLength = fMaxBytes[ch];
        }
        std::cout << "Written to disk " << fileName << " " << byteLength/1024/1024/1024 << " GiB (" << byteLength << " B) for channel " << ch << std::flush;
        TStopwatch writeDisk;
        {
            std::ofstream outF(fileName,std::ios::binary | std::ios::out);
            //std::cout << std::hex << fBufferRAM[ch][0] << " " << fBufferRAM[ch][1] << " " << fBufferRAM[ch][2] << " " << fBufferRAM[ch][3] << std::endl;
            writeDisk.Start();
            std::copy_n(reinterpret_cast<char*>(fBufferRAM[ch]), byteLength, std::ostreambuf_iterator<char>(outF));//https://stackoverflow.com/questions/32050780/ostream-iterator-for-binary-output
            writeDisk.Stop();
            std::cout << " in " << writeDisk.RealTime() << " seconds (" << (static_cast<Double_t>(byteLength/1024/1024)/writeDisk.RealTime()) << " MiB/s)." << std::endl;
            outF.close();
        }

        //@todo save meta-data
        //@todo save CLI output --> TMacro?; as well as software versions
        //@todo boost-log
    }
}

void MainWindow::ToggleSurfer()
{
    if(fRAMSize[0] == 0 && fRAMSize[1] == 0)
    {
        fSurferChannel = -1;
        return;
    }
    if(fSurferChannel > 1)
    {
        fSurferChannel = -1;
        return;
    }
    if(fSurferChannel == -1)
    {
        if(fRAMSize[0]>0)
        {
            fSurferChannel = 0;
        }
        else if(fRAMSize[1]>0)
        {
            fSurferChannel = 1;
        }
        else
        {
            fSurferChannel = -1;
            return;
        }
        fSurfer->Open(reinterpret_cast<uint16_t*>(fBufferRAM[fSurferChannel]),fRAMSize[fSurferChannel]/2,32768,8);
    }
    else
    {
        if(fRAMSize[(fSurferChannel+1)%2] > 0)
        {
            fSurferChannel ++;
            fSurferChannel %= 2;
            fSurfer->Open(reinterpret_cast<uint16_t*>(fBufferRAM[fSurferChannel]),fRAMSize[fSurferChannel]/2,32768,8);
        }
        else if(fRAMSize[fSurferChannel] > 0)
        {
            //fSurfer->Open(reinterpret_cast<uint16_t*>(fBufferRAM[fSurferChannel]),fRAMSize[fSurferChannel]/2,32768,8);
        }
        else
        {
            fSurferChannel = -1;
        }
    }
}

void MainWindow::DoClose()
{
    fThCont->Kill();
    delete fThCont;
    fClockUpdater.Stop();
    fSensorUpdater.Stop();
    fRegisterUpdater.Stop();
    delete[] fBufferRAM[0];
    delete[] fBufferRAM[1];
    fSurfer->CloseWindow();
    DeleteWindow();
    std::cout << "Window closed. Terminating." << std::endl;
    gApplication->Terminate();
}

void MainWindow::SetOutputFolder(const TString folder)
{
    fOutFile->ChangeText(folder);
}

void MainWindow::SetupPRBS(const TString pol, const bool dptc, const bool gz)
{
#ifdef ZDTUSB
    if(!fKeysight || !fKeysight->IsOpen())
    {
        std::cout << "Cannot do PRBS test without Keysight" << std::endl;
        return;
    }
#else
    (void)(pol);
#endif
    if(fThCont->IsRunning())
    {
        std::cout << "Cannot setup while running" << std::endl;
        return;
    }
    //Validate input
    if(pol(0,2) != "PN")
    {
        std::cerr << "Unsupported pol type " << pol << std::endl;
        return;
    }
    const int order = std::stoi(pol(2,pol.Length()-2));
    if(order < 3 || order > 32)
    {
        std::cerr << "Out of range pol type " << pol << std::endl;
        return;
    }
    const uint64_t period = static_cast<uint64_t>(pow(2,order) - 1);
    const uint64_t margin = 120;
    const uint64_t bitRate =  100000000;//or is it 100*1024*1024?
    const uint64_t samRate = 2500000000;//or is it 2500*1024*1024?
    ULong64_t bytes = static_cast<ULong64_t>(ceil(static_cast<double>(period + margin*2)*samRate/bitRate*2.));
    const ULong64_t chunking = 512*2;
    const ULong64_t rest = bytes%chunking;
    if(rest>0)
    {
        bytes += chunking - rest;
    }
    fMaxBytes[0] = bytes;
    const ULong64_t gibytes = 4;
    fRisingEdge->SetState(kButtonDown, kFALSE);
    fExtTrigDelay->SetState(kButtonDown, kFALSE);
    fFrequency->Select(1,kFALSE);
    fDecimation->Select(0,kFALSE);
    fBufferSize[0]->GetNumberEntry()->SetIntNumber(static_cast<Long_t>(gibytes));
    fStore[0]->SetState(kButtonDown, kFALSE);
    fContinuous->SetState(kButtonDown, kFALSE);
    fBinary->SetState(kButtonDown, kFALSE);
    if(dptc)
    {
        fCompress[0]->SetState(kButtonDown, kFALSE);
        fSource[0]->Select(2,kFALSE);
    }
    else
    {
        fSource[0]->Select(0,kFALSE);
    }
    fChunkSize[0]->Select(0,kFALSE);
    fWriteCompression->Select(gz ? 1 : 0,kFALSE);
    DoConfigureSampling();
    DoReserveMemory();
}

void MainWindow::AcquirePRBS()
{
    if(fThCont->IsRunning())
    {
        std::cout << "Cannot acquire while running" << std::endl;
        //return;
    }
    DoArm();//important that it goes before, to clean oupt trash from previous measurements if overflow, as it resets pointers in DMA
    DoStartDAQ();
    fStarter.Connect("Timeout()", "MainWindow", this, "DoKeysightTrigger()");
    fStarter.Start(kStartTimeout,kTRUE);//consider replacing with TTimer::SingleShot();
}

void MainWindow::SetupADCTest(const int pattern, const bool dptc, const bool gz)
{
    fFrequency->Select(1,kFALSE);
    fDecimation->Select(0,kFALSE);
    fBufferSize[0]->GetNumberEntry()->SetIntNumber(static_cast<Long_t>(1));
    fTestMode->Select(pattern,kFALSE);
    fStore[0]->SetState(kButtonDown, kFALSE);
    fContinuous->SetState(kButtonDown, kFALSE);
    fBinary->SetState(kButtonDown, kFALSE);
    if(dptc)
    {
        fCompress[0]->SetState(kButtonDown, kFALSE);
        fSource[0]->Select(2,kFALSE);
    }
    else
    {
        fSource[0]->Select(0,kFALSE);
    }
    fChunkSize[0]->Select(0,kFALSE);
    fWriteCompression->Select(gz ? 1 : 0,kFALSE);
    DoConfigureSampling();
    DoReserveMemory();
}

void MainWindow::AcquireADCTest()
{
    if(fThCont->IsRunning())
    {
        std::cout << "Cannot acquire while running" << std::endl;
        //return;
    }
    DoArm();//important that it goes before, to clean oupt trash from previous measurements if overflow, as it resets pointers in DMA
    DoStartDAQ();
    fStarter.Connect("Timeout()", "MainWindow", this, "DoSoftwareTrigger()");
    fStarter.Start(kStartTimeout,kTRUE);
}
