/**
 * @file
 * @brief PulseSurfer implementation
 */
#include <iostream>
#include <fstream>
#pragma GCC diagnostic ignored "-Wold-style-cast"
#pragma GCC diagnostic ignored "-Wzero-as-null-pointer-constant"
#pragma GCC diagnostic ignored "-Weffc++"
#pragma GCC diagnostic ignored "-Wsign-conversion"
#pragma GCC diagnostic ignored "-Wconversion"
#pragma GCC diagnostic ignored "-Wfloat-conversion"
#pragma GCC diagnostic ignored "-Wcast-qual"
#include "PulseSurfer.hpp"
#include "TStyle.h"
#include "TROOT.h"
#include "TClass.h"
#include "TMath.h"
#include "KeySymbols.h"
#include "TGInputDialog.h"
#include "TGDoubleSlider.h"
#include "TG3DLine.h"
#include "TPaveText.h"
#include "TGFileDialog.h"
#include "TString.h"
#include "TCanvas.h"
#include "TRootEmbeddedCanvas.h"
#include "TH1I.h"
#include "TH2I.h"
#include "TGSlider.h"
#include "TGNumberEntry.h"
#include "TGStatusBar.h"
#include "TGComboBox.h"
#include "TGLabel.h"
#include "TSystem.h"
#include "TTimer.h"
#pragma GCC diagnostic warning "-Wold-style-cast"
#pragma GCC diagnostic warning "-Wzero-as-null-pointer-constant"
#pragma GCC diagnostic warning "-Weffc++"
#pragma GCC diagnostic warning "-Wsign-conversion"
#pragma GCC diagnostic warning "-Wconversion"
#pragma GCC diagnostic warning "-Wfloat-conversion"
#pragma GCC diagnostic warning "-Wcast-qual"

#include <algorithm>    // std::min_element, std::max_element
using std::cout;
using std::endl;
using std::ifstream;

//const UInt_t kWidth=1200;           ///<Canvas size width
//const UInt_t kHeight=kWidth*0.75;   ///<Canvas size height
const char* kCanvasName { "c_pulse" }; ///< Name of main canvas
const char* kHistName { "h_pulse" };   ///< Name of pulse histogram
const char* kHistNameT { "h_fir" };    ///< Name of pulse histogram
const char* kHistName2D { "h2_pulse" };///< Name of 2D pulse histogram

PulseSurfer::PulseSurfer(const TString baseDir) :
TGMainFrame(gClient->GetRoot(), 100, 100),
fQObject(),
fBaseDir(baseDir),
fFileName(""),
fFormat(FileFormat()),
fBytesPerPulse(static_cast<UShort_t>(fFormat.samplesPerPulse*fFormat.bytesPerSample)),
fTotalEvents(0),
fCurrentEvent(0),
fSubtractBaseline(kFALSE),
fPersistency(kFALSE),
fFIR(kFALSE),
fRAM(nullptr),
fFile(),
fRawWaveform(nullptr),
fADC(nullptr),
fCanvas(nullptr),
fHist(nullptr),
fHistT(nullptr),
fHist2D(nullptr),
fStatusBar(nullptr),
fFileFrame(nullptr),
fOpenIcon(nullptr),
fReloadIcon(nullptr),
fFileDialog(nullptr),
fFileEntry(nullptr),
fSamplesPerPulseBox(nullptr),
fPreTriggerBox(nullptr),
fOverflowBox(nullptr),
fBytesPerSampleBox(nullptr),
fSettingsFrame(nullptr),
fSubtractButton(nullptr),
fPersistencyButton(nullptr),
fFIRButton(nullptr),
fEventSlider(nullptr),
fForwardIcon(nullptr),
fRewindIcon(nullptr),
fStopIcon(nullptr),
fTimer(nullptr)
{
    //Set automatic cleanup
    SetCleanup(kDeepCleanup);
    DontCallClose();

    //Create Canvas
    fCanvas = new TRootEmbeddedCanvas(kCanvasName,this,1200,750);
    fCanvas->SetBit(kCannotPick);
    fCanvas->GetCanvas()->Connect("ProcessedEvent(Int_t,Int_t,Int_t,TObject*)", "PulseSurfer",
               this, "CaptureEvent(Int_t,Int_t,Int_t,TObject*)");
    AddFrame(fCanvas, new TGLayoutHints(kLHintsExpandX | kLHintsExpandY, 2, 2, 2, 2));

    //Event status bar below Canvas
    Int_t parts[] = {45, 15, 10, 30};
    fStatusBar = new TGStatusBar(this, 2, 2, kVerticalFrame);
    fStatusBar->SetParts(parts, 4);
    fStatusBar->Draw3DCorner(kFALSE);
    AddFrame(fStatusBar, new TGLayoutHints(kLHintsExpandX, 2, 2, 2, 2));

    //File dialog text entry below Canvas and event status bar
    /* This is the idea:
    ------------------------ fileFrame -----------------------------
    |                                                              |
    |  Open file icon    Filename entry box     Event status bar   |
    |                                                              |
    ----------------------------------------------------------------
    */
    // Create horizontal group frame
    fFileFrame = new TGHorizontalFrame(this,400,25);

    // Create now the open file picture button and add it to the horizontal group frame
    fOpenIcon = new TGPictureButton(fFileFrame,gClient->GetPicture("bld_open.png"));
    fOpenIcon->SetToolTipText("Open file browser");
    fOpenIcon->Connect("Clicked()", "PulseSurfer", this, "OpenFileBrowser()");
    fFileFrame->AddFrame(fOpenIcon, new TGLayoutHints(kLHintsNormal,2,2,1,1));//padleft, padright, padtop, padbottom

    // Create now the reload file picture button and add it to the horizontal group frame
    fReloadIcon = new TGPictureButton(fFileFrame,gClient->GetPicture("bld_replace.xpm"));
    fReloadIcon->SetToolTipText("Reload opened file");
    fReloadIcon->Connect("Clicked()", "PulseSurfer", this, "ReloadFile()");
    fFileFrame->AddFrame(fReloadIcon, new TGLayoutHints(kLHintsNormal,2,2,1,1));//padleft, padright, padtop, padbottom

    // Create now a text entry box inside the horizontal group frame and push it to the left.
    fFileEntry = new TGTextEntry(fFileFrame);
    fFileEntry->SetEnabled(kFALSE);
    fFileEntry->SetToolTipText("Binary file name");
    fFileEntry->Connect("TextChanged(char*)", "PulseSurfer", this, "OpenFromGUI()");
    fFileFrame->AddFrame(fFileEntry, new TGLayoutHints(kLHintsExpandX,2,2,1,1));

    TGLabel* fLabel;
    //Combo boxes to select file format parameters
    fLabel = new TGLabel(fFileFrame,"SP"); fLabel->ChangeOptions(fLabel->GetOptions() | kFixedSize); fLabel->Resize(25, 20); fFileFrame->AddFrame(fLabel, new TGLayoutHints(kLHintsNormal,2,2,1,1));
    fSamplesPerPulseBox = new TGComboBox(fFileFrame, "4096");
    fSamplesPerPulseBox->EnableTextInput(kFALSE);
    fSamplesPerPulseBox->Resize(60,22);
    fSamplesPerPulseBox->AddEntry("32",0);
    fSamplesPerPulseBox->AddEntry("64",1);
    fSamplesPerPulseBox->AddEntry("128",2);
    fSamplesPerPulseBox->AddEntry("256",3);
    fSamplesPerPulseBox->AddEntry("512",4);
    fSamplesPerPulseBox->AddEntry("1024",5);
    fSamplesPerPulseBox->AddEntry("2048",6);
    fSamplesPerPulseBox->AddEntry("4096",7);
    fSamplesPerPulseBox->AddEntry("8192",8);
    fSamplesPerPulseBox->AddEntry("16384",9);
    fSamplesPerPulseBox->AddEntry("32768",10);
    fSamplesPerPulseBox->Select(0,kFALSE);
    fSamplesPerPulseBox->Connect("Selected(int)", "PulseSurfer", this, "UpdateSamplesPerPulse()");
    fFileFrame->AddFrame(fSamplesPerPulseBox, new TGLayoutHints(kLHintsNormal,2,2,1,1));
    fLabel = new TGLabel(fFileFrame,"PT"); fLabel->ChangeOptions(fLabel->GetOptions() | kFixedSize); fLabel->Resize(25, 20); fFileFrame->AddFrame(fLabel, new TGLayoutHints(kLHintsNormal,2,2,1,1));
    fPreTriggerBox = new TGComboBox(fFileFrame, "613");
    fPreTriggerBox->EnableTextInput(kFALSE);
    fPreTriggerBox->Resize(60,22);
    fPreTriggerBox->AddEntry("8",0);
    fPreTriggerBox->AddEntry("613",1);
    fPreTriggerBox->AddEntry("2000",2);
    fPreTriggerBox->Select(0,kFALSE);
    fPreTriggerBox->Connect("Selected(int)", "PulseSurfer", this, "UpdatePreTrigger()");
    fFileFrame->AddFrame(fPreTriggerBox, new TGLayoutHints(kLHintsNormal,2,2,1,1));
    fLabel = new TGLabel(fFileFrame,"Ov"); fLabel->ChangeOptions(fLabel->GetOptions() | kFixedSize); fLabel->Resize(25, 20); fFileFrame->AddFrame(fLabel, new TGLayoutHints(kLHintsNormal,2,2,1,1));
    fOverflowBox = new TGComboBox(fFileFrame, "32767");
    fOverflowBox->EnableTextInput(kFALSE);
    fOverflowBox->Resize(60,22);
    fOverflowBox->AddEntry("32767",0);
    fOverflowBox->AddEntry("16383",1);
    fOverflowBox->Select(0,kFALSE);
    fOverflowBox->Connect("Selected(int)", "PulseSurfer", this, "UpdateOverflow()");
    fFileFrame->AddFrame(fOverflowBox, new TGLayoutHints(kLHintsNormal,2,2,1,1));
    fLabel = new TGLabel(fFileFrame,"By"); fLabel->ChangeOptions(fLabel->GetOptions() | kFixedSize); fLabel->Resize(25, 20); fFileFrame->AddFrame(fLabel, new TGLayoutHints(kLHintsNormal,2,2,1,1));
    fBytesPerSampleBox = new TGComboBox(fFileFrame);//not editable
    fBytesPerSampleBox->EnableTextInput(kFALSE);
    fBytesPerSampleBox->Resize(60,22);
    fBytesPerSampleBox->AddEntry("2",0);
    fBytesPerSampleBox->Select(0,kFALSE);
    fBytesPerSampleBox->SetEnabled(kFALSE);
    fBytesPerSampleBox->Connect("Selected(int)", "PulseSurfer", this, "OpenFromGUI()");
    fFileFrame->AddFrame(fBytesPerSampleBox, new TGLayoutHints(kLHintsNormal,2,2,1,1));
    AddFrame(fFileFrame, new TGLayoutHints(kLHintsExpandX, 2, 2, 2, 2));

    //Horizontal bar below file frame bar
    fSettingsFrame = new TGHorizontalFrame(this,400,25);

    //First item: check button for baseline subtraction
    fLabel = new TGLabel(fSettingsFrame,"BS"); fLabel->ChangeOptions(fLabel->GetOptions() | kFixedSize); fLabel->Resize(25, 20); fSettingsFrame->AddFrame(fLabel, new TGLayoutHints(kLHintsNormal,2,2,1,1));
    fSubtractButton = new TGCheckButton(fSettingsFrame, "",-1);
    fSubtractButton->SetToolTipText("Substract baseline");
    fSubtractButton->Connect("Toggled(Bool_t)","PulseSurfer",this,"SubtractBaseline(Bool_t)");
    fSettingsFrame->AddFrame(fSubtractButton,new TGLayoutHints(kLHintsNormal, 2, 2, 1, 1));

    //Separator
    fSettingsFrame->AddFrame(new TGVertical3DLine(fSettingsFrame,1,22), new TGLayoutHints(kLHintsNormal, 2, 2, 1, 1));

    //Second item: persistency setting
    fLabel = new TGLabel(fSettingsFrame,"2D"); fLabel->ChangeOptions(fLabel->GetOptions() | kFixedSize); fLabel->Resize(25, 20); fSettingsFrame->AddFrame(fLabel, new TGLayoutHints(kLHintsNormal,2,2,1,1));
    fPersistencyButton = new TGCheckButton(fSettingsFrame, "",-1);
    fPersistencyButton->SetToolTipText("Accumulate traces 2D");
    fPersistencyButton->Connect("Toggled(Bool_t)","PulseSurfer",this,"SetPersistency(Bool_t)");
    fSettingsFrame->AddFrame(fPersistencyButton,new TGLayoutHints(kLHintsNormal, 2, 2, 1, 1));

    //Separator
    fSettingsFrame->AddFrame(new TGVertical3DLine(fSettingsFrame,1,22), new TGLayoutHints(kLHintsNormal, 2, 2, 1, 1));

    //Third item: FIR filter
    fLabel = new TGLabel(fSettingsFrame,"FIR"); fLabel->ChangeOptions(fLabel->GetOptions() | kFixedSize); fLabel->Resize(25, 20); fSettingsFrame->AddFrame(fLabel, new TGLayoutHints(kLHintsNormal,2,2,1,1));
    fFIRButton = new TGCheckButton(fSettingsFrame, "",-1);
    fFIRButton->SetToolTipText("Trapezoidal FIR filter");
    fFIRButton->Connect("Toggled(Bool_t)","PulseSurfer",this,"SetFIR(Bool_t)");
    fSettingsFrame->AddFrame(fFIRButton,new TGLayoutHints(kLHintsNormal, 2, 2, 1, 1));

    //Separator
    fSettingsFrame->AddFrame(new TGVertical3DLine(fSettingsFrame,1,22), new TGLayoutHints(kLHintsNormal, 2, 2, 1, 1));

    //Fourth item: slider for event navigation
    fLabel = new TGLabel(fSettingsFrame,"EV"); fLabel->ChangeOptions(fLabel->GetOptions() | kFixedSize); fLabel->Resize(25, 20); fSettingsFrame->AddFrame(fLabel, new TGLayoutHints(kLHintsNormal,2,2,1,1));
    fEventSlider = new TextedHSlider(fSettingsFrame, 190, kDoubleScaleBoth, 0,
                                   kHorizontalFrame,
                                   fSettingsFrame->GetBackground(),
                                   0, 1, 0, kFALSE);
    fEventSlider->GetNumberEntry()->Resize(75,22);
    fEventSlider->GetSlider()->Connect("PositionChanged(Int_t)","PulseSurfer",this,"ReadPulseEmit(Int_t)");
    Connect("ReadPulse(ULong64_t,Bool_t)", "TCanvas", fCanvas->GetCanvas(), "Update()");
    fSettingsFrame->AddFrame(fEventSlider->GetSlider(), new TGLayoutHints(kLHintsTop | kLHintsExpandX, 2, 2, 1, 1));
    fSettingsFrame->AddFrame(fEventSlider->GetNumberEntry(), new TGLayoutHints(kLHintsTop , 2, 2, 1, 1));

    //Fifth item: event player
    fRewindIcon = new TGPictureButton(fSettingsFrame,gClient->GetPicture("first_t.xpm"));
    fRewindIcon->SetToolTipText("Backward autoplay");
    fRewindIcon->Resize(22,22);
    fRewindIcon->Connect("Clicked()", "PulseSurfer", this, "Rewind()");
    fSettingsFrame->AddFrame(fRewindIcon, new TGLayoutHints(kLHintsNormal,2,2,1,1));//padleft, padright, padtop, padbottom
    fForwardIcon = new TGPictureButton(fSettingsFrame,gClient->GetPicture("last_t.xpm"));
    fForwardIcon->SetToolTipText("Forward autoplay");
    fForwardIcon->Resize(22,22);
    fForwardIcon->Connect("Clicked()", "PulseSurfer", this, "Play()");
    fSettingsFrame->AddFrame(fForwardIcon, new TGLayoutHints(kLHintsNormal,2,2,1,1));//padleft, padright, padtop, padbottom
    fStopIcon = new TGPictureButton(fSettingsFrame,gClient->GetPicture("ed_interrupt.png"));
    fStopIcon->SetToolTipText("Stop autoplay");
    fStopIcon->Resize(22,22);
    fStopIcon->Connect("Clicked()", "PulseSurfer", this, "Stop()");
    fSettingsFrame->AddFrame(fStopIcon, new TGLayoutHints(kLHintsNormal,2,2,1,1));//padleft, padright, padtop, padbottom

    //Timer:
    fTimer = new TTimer(50);
    fTimer->Connect("Timeout()", "PulseSurfer", this, "TimerUpdate()");

    //Add horizontal bar to main frame
    AddFrame(fSettingsFrame, new TGLayoutHints(kLHintsExpandX, 2, 2, 2, 2));

    //Disable until first open
    Freeze();

    //Map all windows
    SetWindowName("Scroll mouse wheel on canvas, or press n / b for next / back pulses. Or use slider. Fix your zoom, right click axes, SetRangeUser.");
    MapSubwindows();
    Resize(GetDefaultSize());
    MapWindow();

    //Set default graphics format
    //fCanvas->GetCanvas()->SetSupportGL(true);//NEEDED for fast drawing.
    gStyle->SetOptStat(0);
}

Bool_t PulseSurfer::Open(const TString fileName, const FileFormat format)
{
    SetRAM(nullptr);
    //Check byte size
    if(format.bytesPerSample!=2)
    {
        cout << "Error in PulseSurfer::Open: just 2 bytes per sample supported up to now, but you asked for " << format.bytesPerSample << ". Class should be changed or templated if anything else is wanted." << endl;
        return kFALSE;
    }

    //Check the total number of events of this file
    const ULong64_t evts = CheckFile(fileName,static_cast<UShort_t>(format.bytesPerSample*format.samplesPerPulse));
    if(evts == 0)
    {
        cout << "Error in PulseSurfer::Open: no events are found in this file." << endl;
        return kFALSE;
    }

    //Check preTrigger range
    if(format.preTrigger>=format.samplesPerPulse)
    {
        cout << "Error in PulseSurfer::Open: preTrigger value (" << format.preTrigger << ") >= samples per pulse (" << format.samplesPerPulse << ")." << endl;
        return kFALSE;
    }

    Stop();//stop timer if running
    fCanvas->GetCanvas()->Update();//Draw everything that is missing

    fFileName=fileName;
    //Set base directory
    fBaseDir = fFileName;
    const Int_t pos = fBaseDir.Last('/');
    const Int_t len = fBaseDir.Length();
    if(pos==-1) fBaseDir="";
    else fBaseDir.Remove(pos,len-pos);
    fFormat = format;
    fBytesPerPulse=static_cast<UInt_t>(fFormat.bytesPerSample*fFormat.samplesPerPulse);
    fTotalEvents= evts;
    fCurrentEvent = 0;
    fSubtractButton->SetEnabled(kTRUE);
    fPersistencyButton->SetEnabled(kTRUE);
    fFIRButton->SetEnabled(kTRUE);
    fEventSlider->GetSlider()->SetRange(0,static_cast<Int_t>(fTotalEvents-1));
    fEventSlider->GetNumberEntry()->SetLimitValues(0,static_cast<Double_t>(fTotalEvents-1));
    fEventSlider->GetSlider()->SetPosition(static_cast<Int_t>(fCurrentEvent));
    fEventSlider->GetNumberEntry()->SetIntNumber(static_cast<Long64_t>(fCurrentEvent));
    fEventSlider->GetSlider()->SetEnabled(kTRUE);
    fEventSlider->GetNumberEntry()->SetState(kTRUE);
    fRewindIcon->SetEnabled(kTRUE); fRewindIcon->SetState(kButtonUp);
    fForwardIcon->SetEnabled(kTRUE); fForwardIcon->SetState(kButtonUp);
    fStopIcon->SetEnabled(kTRUE); fStopIcon->SetState(kButtonDown);


    //Update GUI boxes but don't emit
    fFileEntry->SetText(fFileName,kFALSE);
    fSamplesPerPulseBox->Select(fSamplesPerPulseBox->FindEntry(std::to_string(static_cast<int>(fFormat.samplesPerPulse)).c_str())->EntryId(),kFALSE);
    fPreTriggerBox->Select(fPreTriggerBox->FindEntry(std::to_string(static_cast<int>(fFormat.preTrigger)).c_str())->EntryId(),kFALSE);
    fOverflowBox->Select(fOverflowBox->FindEntry(std::to_string(fFormat.overflow).c_str())->EntryId(),kFALSE);
    fBytesPerSampleBox->Select(fBytesPerSampleBox->FindEntry(std::to_string(static_cast<int>(fFormat.bytesPerSample)).c_str())->EntryId(),kFALSE);

    //Close previous things, if existing
    fFile.close();
    if(gROOT->FindObject(kHistName)) delete fHist;
    fHist=nullptr;
    if(gROOT->FindObject(kHistNameT)) delete fHistT;
    fHistT=nullptr;
    if(gROOT->FindObject(kHistName2D)) delete fHist2D;
    fHist2D=nullptr;
    delete[] fRawWaveform;
    delete[] fADC;

    //Open file streaming
    fFile.open(fFileName,std::ios::binary);
    fRawWaveform = new Char_t[fBytesPerPulse];
    fADC = new Short_t[fFormat.samplesPerPulse];

    RecreateHistogram();

    //Read already first pulse
    const Bool_t res = ReadNextPulse();
    fCanvas->GetCanvas()->Update();

    SetPersistency(fPersistency);//recreates 2D histo (after pulse is plotted!, in case no set user range zoom was set)

    return res;
}

void PulseSurfer::SetRAM(uint16_t* const buffer)
{
    fRAM = buffer;
}

Bool_t PulseSurfer::Open(uint16_t* const bufferRAM, const ULong64_t samples, const FileFormat format)
{
    fFile.close();

    //Check byte size
    if(format.bytesPerSample!=2)
    {
        cout << "Error in PulseSurfer::Open: just 2 bytes per sample supported up to now, but you asked for " << format.bytesPerSample << ". Class should be changed or templated if anything else is wanted." << endl;
        return kFALSE;
    }

    //Check preTrigger range
    if(format.preTrigger>=format.samplesPerPulse)
    {
        cout << "Error in PulseSurfer::Open: preTrigger value (" << format.preTrigger << ") >= samples per pulse (" << format.samplesPerPulse << ")." << endl;
        return kFALSE;
    }

    //Check samples per pulse
    if(samples%format.samplesPerPulse==0)
    {
        cout << "Reading from RAM: " << samples/format.samplesPerPulse << " events." << endl;
    }
    else
    {
        cout << "RAM size is corrupt, size not divisible by " << format.samplesPerPulse << ", rest is " << samples%format.samplesPerPulse << "!" << endl;
        return kFALSE;
    }

    SetRAM(bufferRAM);

    Stop();//stop timer if running
    fCanvas->GetCanvas()->Update();//Draw everything that is missing

    fFileName="";
    //Set base directory
    fBaseDir = "";
    fFormat = format;
    fBytesPerPulse=static_cast<UInt_t>(fFormat.bytesPerSample*fFormat.samplesPerPulse);
    fTotalEvents= samples/format.samplesPerPulse;
    fCurrentEvent = 0;
    fSubtractButton->SetEnabled(kTRUE);
    fPersistencyButton->SetEnabled(kTRUE);
    fFIRButton->SetEnabled(kTRUE);
    fEventSlider->GetSlider()->SetRange(0,static_cast<Int_t>(fTotalEvents-1));
    fEventSlider->GetNumberEntry()->SetLimitValues(0,static_cast<Double_t>(fTotalEvents-1));
    fEventSlider->GetSlider()->SetPosition(static_cast<Int_t>(fCurrentEvent));
    fEventSlider->GetNumberEntry()->SetIntNumber(static_cast<Long64_t>(fCurrentEvent));
    fEventSlider->GetSlider()->SetEnabled(kTRUE);
    fEventSlider->GetNumberEntry()->SetState(kTRUE);
    fRewindIcon->SetEnabled(kTRUE); fRewindIcon->SetState(kButtonUp);
    fForwardIcon->SetEnabled(kTRUE); fForwardIcon->SetState(kButtonUp);
    fStopIcon->SetEnabled(kTRUE); fStopIcon->SetState(kButtonDown);


    //Update GUI boxes but don't emit
    fFileEntry->SetText(fFileName,kFALSE);
    fSamplesPerPulseBox->Select(fSamplesPerPulseBox->FindEntry(std::to_string(static_cast<int>(fFormat.samplesPerPulse)).c_str())->EntryId(),kFALSE);
    fPreTriggerBox->Select(fPreTriggerBox->FindEntry(std::to_string(static_cast<int>(fFormat.preTrigger)).c_str())->EntryId(),kFALSE);
    fOverflowBox->Select(fOverflowBox->FindEntry(std::to_string(fFormat.overflow).c_str())->EntryId(),kFALSE);
    fBytesPerSampleBox->Select(fBytesPerSampleBox->FindEntry(std::to_string(static_cast<int>(fFormat.bytesPerSample)).c_str())->EntryId(),kFALSE);

    //Close previous things, if existing
    fFile.close();
    if(gROOT->FindObject(kHistName)) delete fHist;
    fHist=nullptr;
    if(gROOT->FindObject(kHistNameT)) delete fHistT;
    fHistT=nullptr;
    if(gROOT->FindObject(kHistName2D)) delete fHist2D;
    fHist2D=nullptr;
    delete[] fRawWaveform;
    delete[] fADC;

    fRawWaveform = nullptr;
    fADC = nullptr;

    RecreateHistogram();

    //Read already first pulse
    const Bool_t res = ReadNextPulse();
    fCanvas->GetCanvas()->Update();

    SetPersistency(fPersistency);//recreates 2D histo (after pulse is plotted!, in case no set user range zoom was set)

    return res;
}

Bool_t PulseSurfer::Open(const TString fileName, const UShort_t samplesPerPulse, const UShort_t preTrigger, const Short_t overflow, const UShort_t bytesPerSample)
{
    FileFormat ff;
    ff.samplesPerPulse=samplesPerPulse;
    ff.preTrigger=preTrigger;
    ff.overflow=overflow;
    ff.bytesPerSample=bytesPerSample;
    return Open(fileName,ff);
}

Bool_t PulseSurfer::Open(uint16_t* const ramBuffer, const ULong64_t samples, const UShort_t samplesPerPulse, const UShort_t preTrigger, const Short_t overflow, const UShort_t bytesPerSample)
{
    FileFormat ff;
    ff.samplesPerPulse=samplesPerPulse;
    ff.preTrigger=preTrigger;
    ff.overflow=overflow;
    ff.bytesPerSample=bytesPerSample;
    return Open(ramBuffer,samples,ff);
}

void PulseSurfer::Freeze()
{
    fSubtractButton->SetEnabled(kFALSE);
    fPersistencyButton->SetEnabled(kFALSE);
    fFIRButton->SetEnabled(kFALSE);
    fEventSlider->GetSlider()->SetEnabled(kFALSE);
    fEventSlider->GetNumberEntry()->SetState(kFALSE);
    fRewindIcon->SetEnabled(kFALSE);
    fForwardIcon->SetEnabled(kFALSE);
    fStopIcon->SetEnabled(kFALSE);
}

void PulseSurfer::ReloadFile()
{
    OpenFromGUI();
}

void PulseSurfer::OpenFileBrowser()
{
    TGFileInfo fi;
    const char *inputFileTypes[] =
    { "Binary files",   "*.[bd]in",
      //~ "ROOT files",     "*.root",
                nullptr,               nullptr };
    fi.fFileTypes = inputFileTypes;
    if(fBaseDir=="")
    {
        delete[] fi.fIniDir;
        fi.fIniDir    = StrDup(gSystem->pwd());
    }
    else
    {
        delete[] fi.fIniDir;
        fi.fIniDir    = StrDup(fBaseDir);
    }
    if(fFileName!="")
    {
        delete[] fi.fFilename;
        fi.fFilename    = StrDup(fFileName);
    }

    fFileDialog = new TGFileDialog(gClient->GetRoot(), this, kFDOpen, &fi);
    fBaseDir = fi.fIniDir;
    fFileEntry->SetText(fi.fFilename);
    return;
}

Bool_t PulseSurfer::OpenFromGUI()
{
    if(!fSamplesPerPulseBox->GetSelectedEntry() || !fPreTriggerBox->GetSelectedEntry() || !fOverflowBox->GetSelectedEntry() || !fBytesPerSampleBox->GetSelectedEntry() || static_cast<TString>(fFileEntry->GetText())=="") return kFALSE;//For safety, in case nothing is selected

    //http://osdir.com/ml/lang.c++.root/2005-11/msg00067.html
    FileFormat ff;
    ff.samplesPerPulse=static_cast<UShort_t>(atoi(static_cast<TGTextLBEntry *>(fSamplesPerPulseBox->GetSelectedEntry())->GetTitle()));
    ff.preTrigger=static_cast<UShort_t>(atoi(static_cast<TGTextLBEntry *>(fPreTriggerBox->GetSelectedEntry())->GetTitle()));
    ff.overflow=static_cast<Short_t>(atoi(static_cast<TGTextLBEntry *>(fOverflowBox->GetSelectedEntry())->GetTitle()));
    ff.bytesPerSample=static_cast<UShort_t>(atoi(static_cast<TGTextLBEntry *>(fBytesPerSampleBox->GetSelectedEntry())->GetTitle()));
    return Open(fFileEntry->GetText(),ff);
}

void PulseSurfer::RecreateHistogram()
{
    fHist = new TH1I(kHistName,"",fFormat.samplesPerPulse,0,fFormat.samplesPerPulse);
    fHist->GetXaxis()->SetTitle("Sample point");
    fHist->SetLineColor(kBlack);
    SubtractBaseline(fSubtractBaseline);//sets y-axis title
    fCanvas->GetCanvas()->cd();
    fHist->Draw();
    fCanvas->GetCanvas()->Update();
}

PulseSurfer::~PulseSurfer()
{

}

ULong64_t PulseSurfer::CheckFile(const TString fileName, const UShort_t bytesPerPulse)
{
    ifstream f(fileName,std::ios::binary);
    if (!f.is_open() || f.fail() || !f.good())
    {
        cout  << "Error in PulseSurfer::CheckFile: file " << fileName << " could not be opened. Exiting! " << endl;
        f.close();
        return 0;
    }

    //File size
    f.seekg(0,std::ios::end);
    const ULong64_t fileSize= static_cast<ULong64_t>(f.tellg());

    cout << "\tNote in PulseSurfer::CheckFile: The file " << fileName << " contains " << fileSize << " bytes";
    if(fileSize%bytesPerPulse==0)
    {
        cout << "; " << fileSize/bytesPerPulse << " events." << endl;
        f.close();
        return fileSize/bytesPerPulse;
    }
    else
    {
        cout << ". File is corrupt, size not divisible by " << bytesPerPulse << ", rest is " << fileSize%bytesPerPulse << "!" << endl;
        f.close();
        return 0;
    }
}

Bool_t PulseSurfer::ReadNextPulse()
{
    if(!fFile.is_open() && !fRAM)
    {
        cout << "Error in PulseSurfer::ReadNextPulse: file / RAM is not open." << endl;
        return kFALSE;
    }
    if(!fHist)
    {
        cout << "Error in PulseSurfer::ReadNextPulse: hist does not exist." << endl;
        return kFALSE;
    }

    if(fCurrentEvent==fTotalEvents)
    {
        //~ cout << "Error in PulseSurfer::ReadNextPulse: reached end of file." << endl;
        return kFALSE;
    }
    //Check the number of bytes
    if(fFormat.bytesPerSample!=2)
    {
        cout << "Error in PulseSurfer::ReadNextPulse: just 2 bytes per sample supported up to now, but you asked for " << fFormat.bytesPerSample << ". Class should be changed or templated if anything else is wanted." << endl;
        return kFALSE;
    }
    const ULong64_t offset = fCurrentEvent*static_cast<ULong64_t>(fFormat.samplesPerPulse);

    //~ cout << "\tNote in PulseSurfer::ReadNextPulse: pulse " << fCurrentEvent << " is read.\r" << flush;
    /** Reset baseline counting array */
    if(fSubtractBaseline) std::fill(fFrequence, fFrequence+65536, 0);//reset all to zero

    if(fRAM)
    {
        /** Fill pulse ADC array and count frequence for getting the MPV baseline */
        if(fSubtractBaseline)
        {
            for(UShort_t s=0;s<fFormat.preTrigger;++s)
            {
                fFrequence[fRAM[offset+s]]++;
            }
        }
    }
    else
    {
        fFile.read(fRawWaveform,fBytesPerPulse);

        /** Fill pulse ADC array and count frequence for getting the MPV baseline */
        for(UShort_t s=0;s<fFormat.samplesPerPulse;++s)
        {
            fADC[s]=static_cast<Short_t>((static_cast<Short_t>(static_cast<UChar_t>(fRawWaveform[s*fFormat.bytesPerSample+1]))<<8) | (static_cast<Short_t>(static_cast<UChar_t>(fRawWaveform[s*fFormat.bytesPerSample]))));
            if(fSubtractBaseline && s<fFormat.preTrigger) fFrequence[UShort_t(fADC[s])]++;
        }
    }

    //~ if(!gROOT->FindObject(kCanvasName)) RecreateCanvas();//someone erased it
    //~ if(!gROOT->FindObject(kHistName)) RecreateHistogram();//someone erased it

    fHist->SetTitle((std::to_string(fCurrentEvent)+" / "+std::to_string(fTotalEvents-1)).c_str());
    if(fHistT) fHistT->SetTitle((std::to_string(fCurrentEvent)+" / "+std::to_string(fTotalEvents-1)).c_str());
    if(fHist2D) fHist2D->SetTitle((std::to_string(fCurrentEvent)+" / "+std::to_string(fTotalEvents-1)).c_str());
    Short_t baseline=0;
    if(fSubtractBaseline)
    {
        fFrequence[fFormat.overflow]=0;//Avoid having the baseline exactly at the overflow position
        baseline = Short_t(std::distance(fFrequence,std::max_element(fFrequence,fFrequence+65536)));
    }
    if(fRAM)
    {
        for(UShort_t s=0;s<fFormat.samplesPerPulse;++s)
        {
            fHist->SetBinContent(s+1,fRAM[offset + s]-baseline);
            if(fHist2D) fHist2D->Fill(s,fRAM[offset + s]-baseline);
        }
    }
    else
    {
        for(UShort_t s=0;s<fFormat.samplesPerPulse;++s)
        {
            fHist->SetBinContent(s+1,fADC[s]-baseline);
            if(fHist2D) fHist2D->Fill(s,fADC[s]-baseline);
        }
    }

    if(fHistT && fFIR)
    {
        //const Double_t km = 0.25;
        //const Double_t k0 = 0.50;
        //const Double_t kp = 0.25;

        fHistT->SetBinContent(1,0);
        //~ fHistT->SetBinContent(2,0);
        //~ fHistT->SetBinContent(fFormat.samplesPerPulse,0);
        //~ fHistT->SetBinContent(fFormat.samplesPerPulse-1,0);
        //~ for(UShort_t s=1;s<fFormat.samplesPerPulse-1;++s)
        //~ {
            //~ fHistT->SetBinContent(s+1,k0*(fADC[s]-baseline)+km*(fADC[s-1]-baseline)+kp*(fADC[s+1]-baseline));
        //~ }

        //~ for(UShort_t s=2;s<fFormat.samplesPerPulse-2;++s)
        //~ {
            //~ fHistT->SetBinContent(s+1,k0*(fADC[s]-baseline)+km*(fADC[s-1]-baseline)+kp*(fADC[s+1]-baseline)-(k0*(fADC[s-1]-baseline)+km*(fADC[s-1-1]-baseline)+kp*(fADC[s+1-1]-baseline)));
        //~ }
        const double dRate = exp(-1/8.);
        if(fRAM)
        {
            for(UShort_t s=1;s<fFormat.samplesPerPulse;++s)
            {
                fHistT->SetBinContent(s+1,-(dRate*(fRAM[offset + s-1]-baseline)-(fRAM[offset + s]-baseline)));
            }
        }
        else
        {
            for(UShort_t s=1;s<fFormat.samplesPerPulse;++s)
            {
                fHistT->SetBinContent(s+1,-(dRate*(fADC[s-1]-baseline)-(fADC[s]-baseline)));
            }
        }
    }

    fEventSlider->GetSlider()->SetPosition(static_cast<Int_t>(fCurrentEvent));
    fEventSlider->GetNumberEntry()->SetIntNumber(static_cast<Long64_t>(fCurrentEvent));

    fCurrentEvent++;
    return kTRUE;
}

Bool_t PulseSurfer::ReadPreviousPulse()
{
    if(fCurrentEvent==0 || fCurrentEvent==1)
    {
        //~ cout << "Error in PulseSurfer::ReadPreviousPulse: reached file beginning." << endl;
        return kFALSE;
    }
    fCurrentEvent-=2;
    return ReadPulse(fCurrentEvent,kFALSE);
}

Bool_t PulseSurfer::ReadPulse(const ULong64_t event, const Bool_t emit)
{
    if(event>=fTotalEvents)
    {
        cout << "Error in PulseSurfer::ReadPulse: pulse " << event << " out of range >= " << fTotalEvents << endl;
        if(event==fTotalEvents) cout << "Note that first pulse starts with id 0." << endl;
        return kFALSE;
    }

    if(!fFile.is_open() && !fRAM)
    {
        cout << "Error in PulseSurfer::ReadPulse: file / RAM is not open." << endl;
        return kFALSE;
    }

    if(!fHist)
    {
        cout << "Error in PulseSurfer::ReadPulse: hist does not exist." << endl;
        return kFALSE;
    }

    if(!fRAM)
    {
        fFile.seekg(static_cast<long>(event*fBytesPerPulse),fFile.beg);
    }
    fCurrentEvent=event;
    const Bool_t res = ReadNextPulse();
    if(res && emit) Emit("ReadPulse(ULong64_t,Bool_t)");
    return res;
}

Bool_t PulseSurfer::ReadPulseEmit(const ULong64_t event)
{
    return ReadPulse(event,kTRUE);//This auxiliary function is used because fEventSlider->GetSlider()->Connect("PositionChanged(Int_t)","PulseSurfer",this,"ReadPulse(Int_t,=kTRUE)"); does not work
}

void PulseSurfer::CaptureEvent(Int_t event, Int_t px, Int_t py, TObject *sel)
{
    if(!sel) return;
    //  Writes the event status in the status bar parts
    fStatusBar->SetText(sel->GetTitle(),0);
    fStatusBar->SetText(sel->GetName(),1);
    if (event == kKeyPress)
    {
        fStatusBar->SetText(Form("%c", static_cast<char>(px)),2);
        fStatusBar->SetText(Form("0x%x",                 py) ,3);
    }
    else
    {
        fStatusBar->SetText(Form("%d,%d", px, py),2);
        fStatusBar->SetText(sel->GetObjectInfo(px,py),3);
    }

    if(!fFile.is_open() && !fRAM) return;

    if(event==kWheelUp && !(TString(sel->IsA()->GetName())=="TAxis") && !(TString(sel->IsA()->GetName())=="TH2I"))//not when zooming axis
    {
        ReadNextPulse();
        fCanvas->GetCanvas()->Update();
    }
    else if(event==kWheelDown && !(TString(sel->IsA()->GetName())=="TAxis") && !(TString(sel->IsA()->GetName())=="TH2I"))//not when zooming axis
    {
        ReadPreviousPulse();
        fCanvas->GetCanvas()->Update();
    }
    else if(event == kKeyPress)// Actions in reponse to key press events. Note: arrows cannot be captured, they actually move the cursor.
    {
        //~ Int_t key=px;
        Int_t keysym=py;
        //~ printf("symbol = %c (%x)\n", key, keysym);
        if(keysym==kKey_N || keysym==kKey_n) {
            ReadNextPulse();
            fCanvas->GetCanvas()->Update();
        }
        else if(keysym==kKey_B || keysym==kKey_b) {
            ReadPreviousPulse();
            fCanvas->GetCanvas()->Update();
        }
    }
}

void PulseSurfer::SubtractBaseline(const Bool_t subtract)
{
    fSubtractBaseline=subtract;
    if(!fHist)
    {
        //~ cout << "Error in PulseSurfer::SubtractBaseline: hist not available" << endl;
        return;
    }
    if(fSubtractButton) fSubtractButton->SetDown(fSubtractBaseline,kFALSE);//don't emit
    if(fSubtractBaseline) fHist->GetYaxis()->SetTitle("Channel - Baseline / ADC");
    else fHist->GetYaxis()->SetTitle("Channel / ADC");
    if(fHistT)
    {
        fHistT->GetXaxis()->SetTitle(fHist->GetXaxis()->GetTitle());
        fHistT->GetYaxis()->SetTitle(fHist->GetYaxis()->GetTitle());
    }
    if(fHist2D)
    {
        fHist2D->GetXaxis()->SetTitle(fHist->GetXaxis()->GetTitle());
        fHist2D->GetYaxis()->SetTitle(fHist->GetYaxis()->GetTitle());
    }
}

void PulseSurfer::SetPersistency(const Bool_t persist)
{
    if(!fFile.is_open() && !fRAM)
    {
        cout << "Error in PulseSurfer::SetPersistency: file/RAM is not yet open, call again later." << endl;
        fPersistencyButton->SetDown(fPersistency,kFALSE);//don't emit
        return;
    }

    fPersistency=persist;
    fPersistencyButton->SetDown(fPersistency,kFALSE);//don't emit

    if(!fPersistency)
    {
        if(gROOT->FindObject(kHistName2D)) delete fHist2D;
        fHist2D=nullptr;

        fHist->SetLineWidth(1);
        fCanvas->GetCanvas()->cd();
        fHist->Draw();
        fCanvas->GetCanvas()->Update();
    }
    else
    {
        if(!fHist2D)
        {
            fHist2D = new TH2I(kHistName2D, fHist->GetTitle(),
                                fHist->GetXaxis()->GetLast()-fHist->GetXaxis()->GetFirst()+1,
                                fHist->GetXaxis()->GetBinLowEdge(fHist->GetXaxis()->GetFirst()),
                                fHist->GetXaxis()->GetBinUpEdge(fHist->GetXaxis()->GetLast()),
                                TMath::Max(TMath::Min(TMath::Nint((gPad->GetUymax()-gPad->GetUymin())/50.),1000),200),
                                gPad->GetUymin(),
                                gPad->GetUymax());
            fHist2D->GetXaxis()->SetTitle(fHist->GetXaxis()->GetTitle());
            fHist2D->GetYaxis()->SetTitle(fHist->GetYaxis()->GetTitle());
            fHist2D->SetContour(255);
            fCanvas->GetCanvas()->cd();
            fHist2D->Draw("COLZ");
            fHist->SetLineWidth(2);
            fHist->Draw("same");
            if(fHistT) fHistT->Draw("same");
            fCanvas->GetCanvas()->Update();
        }
    }
}

void PulseSurfer::SetFIR(const Bool_t fir)
{
    fFIR=fir;
    if(fFIRButton) fFIRButton->SetDown(fFIR,kFALSE);//don't emit

    if(!fFIR)
    {
        if(gROOT->FindObject(kHistNameT)) delete fHistT;
        fHistT=nullptr;
    }
    else
    {
        if(!fHistT && fHist)
        {
            fHistT = new TH1F(kHistNameT, fHist->GetTitle(),fFormat.samplesPerPulse,0,fFormat.samplesPerPulse);
            fHistT->GetXaxis()->SetTitle(fHist->GetXaxis()->GetTitle());
            fHistT->GetYaxis()->SetTitle(fHist->GetYaxis()->GetTitle());
            fCanvas->GetCanvas()->cd();
            fHistT->SetLineColor(kRed);
            fHistT->Draw("same");
            fCanvas->GetCanvas()->Update();
        }
    }
}

void PulseSurfer::CloseWindow()
{
   // Called when window is closed via the window manager.
   //delete this;
    delete[] fRawWaveform;
    delete[] fADC;
    fFile.close();
    fRAM = nullptr;
    if(gROOT->FindObject(kHistName)) delete fHist;//we check that user did not close it already
    if(gROOT->FindObject(kHistNameT)) delete fHistT;//we check that user did not close it already
    if(gROOT->FindObject(kHistName2D)) delete fHist2D;//we check that user did not close it already
    delete fTimer;
    delete fEventSlider;
    std::cout << "Closing PulseSurfer." << std::endl;
    DeleteWindow();
}

Bool_t PulseSurfer::UpdateSamplesPerPulse()
{
    if(!fSamplesPerPulseBox->GetSelectedEntry()) return kFALSE;
    Int_t samplesPerPulse = atoi(static_cast<TGTextLBEntry *>(fSamplesPerPulseBox->GetSelectedEntry())->GetTitle());
    if(samplesPerPulse>UShort_t(-1))
    {
        cout << "Error in PulseSurfer::UpdateSamplesPerPulse: value " << samplesPerPulse << " would be truncated to UShort_t, until 65535. Set new value." << endl;
        return kFALSE;
    }
    if(samplesPerPulse<0)
    {
        cout << "Error in PulseSurfer::UpdateSamplesPerPulse: value " << samplesPerPulse << " is negative. Set new value." << endl;
        return kFALSE;
    }
    return OpenFromGUI();
}

Bool_t PulseSurfer::UpdatePreTrigger()
{
    if(!fPreTriggerBox->GetSelectedEntry()) return kFALSE;
    Int_t preTrigger = atoi(static_cast<TGTextLBEntry *>(fPreTriggerBox->GetSelectedEntry())->GetTitle());
    if(preTrigger>UShort_t(-1))
    {
        cout << "Error in PulseSurfer::UpdatePreTrigger: value " << preTrigger << " would be truncated to UShort_t, until 65535. Set new value." << endl;
        return kFALSE;
    }
    if(preTrigger<0)
    {
        cout << "Error in PulseSurfer::UpdatePreTrigger: value " << preTrigger << " is negative. Set new value." << endl;
        return kFALSE;
    }
    preTrigger = UShort_t(preTrigger);
    if((fFile.is_open() || fRAM) && preTrigger>=fFormat.samplesPerPulse)//if file is closed, samplesPerPulse may be not yet set via GUI. Dont worry, will be rechecked when calling Open once samplesPerPulse is set.
    {
        cout << "Error in PulseSurfer::UpdatePreTrigger: value (" << preTrigger << ") >= number of samples per pulse (" << fFormat.samplesPerPulse << "). Set new value." << endl;
        return kFALSE;
    }
    fFormat.preTrigger=static_cast<UShort_t>(preTrigger);
    return kTRUE;
}

Bool_t PulseSurfer::UpdateOverflow()
{
    if(!fOverflowBox->GetSelectedEntry()) return kFALSE;
    Int_t overflow = atoi(static_cast<TGTextLBEntry *>(fOverflowBox->GetSelectedEntry())->GetTitle());
    if(overflow>32767)
    {
        cout << "Error in PulseSurfer::UpdateOverflow: value " << overflow << " would be truncated to Short_t, until 32767." << endl;
        return kFALSE;
    }
    if(overflow<-32768)
    {
        cout << "Error in PulseSurfer::UpdateOverflow: value " << overflow << " would be truncated to Short_t, until -32768." << endl;
        return kFALSE;
    }
    fFormat.overflow=static_cast<Short_t>(overflow);
    return kTRUE;
}

void PulseSurfer::Rewind()
{
    fTimer->Stop();//if it comes from fw play

    if(!fRewindIcon->IsEnabled() || !fForwardIcon->IsEnabled() || !fStopIcon->IsEnabled()) return;

    fRewindIcon->SetState(kButtonDown);
    fForwardIcon->SetState(kButtonUp);
    fStopIcon->SetState(kButtonUp);

    fTimer->Start();
}

void PulseSurfer::Play()
{
    fTimer->Stop();//if it comes from bw play

    if(!fRewindIcon->IsEnabled() || !fForwardIcon->IsEnabled() || !fStopIcon->IsEnabled()) return;

    fRewindIcon->SetState(kButtonUp);
    fForwardIcon->SetState(kButtonDown);
    fStopIcon->SetState(kButtonUp);

    fTimer->Start();
}

void PulseSurfer::Stop()
{
    fTimer->Stop();//for symmetry

    if(!fRewindIcon->IsEnabled() || !fForwardIcon->IsEnabled() || !fStopIcon->IsEnabled()) return;

    fRewindIcon->SetState(kButtonUp);
    fForwardIcon->SetState(kButtonUp);
    fStopIcon->SetState(kButtonDown);

    fTimer->Stop();
}

Bool_t PulseSurfer::TimerUpdate()
{
    if(fStopIcon->IsDown())//Error handling
    {
        Stop();
        return kFALSE;
    }
    else if(fRewindIcon->IsDown() && fForwardIcon->IsDown())//Error handling
    {
        //give it later another chance, don't stop, sometimes when clicking directly from FW to BW this can happen
        return kFALSE;
    }
    else if(fRewindIcon->IsDown())
    {
        const Bool_t res = ReadPreviousPulse();
        if(res) fCanvas->GetCanvas()->Update();
        else Stop();
        return res;
    }
    else if(fForwardIcon->IsDown())
    {
        const Bool_t res = ReadNextPulse();
        if(res) fCanvas->GetCanvas()->Update();
        else Stop();
        return res;
    }
    else//Error handling
    {
        Stop();
        return kFALSE;
    }
}

TextedHSlider::TextedHSlider(const TGWindow* p, UInt_t w, UInt_t type, Int_t id, UInt_t options, Pixel_t back, const Double_t start, const Double_t end, const Double_t position, Bool_t isReal)
{
    if(!isReal)
    {
        fSlider = new TGHSlider(p,w,type,id,options,back);
        fSlider->SetRange(TMath::Nint(start),TMath::Nint(end));
        fSlider->SetPosition(TMath::Nint(position));
        fSlider->Connect("PositionChanged(Int_t)", "TextedHSlider", this, "MoveNumber(Int_t)");

        if(start>=0)
        {
            fNumber = new TGNumberEntry(p,start,TMath::Nint(TMath::Log(end)/TMath::Log(10)), 0, TGNumberFormat::kNESInteger, TGNumberFormat::kNEANonNegative, TGNumberFormat::kNELLimitMinMax, start, end);
        }
        else
        {
            fNumber = new TGNumberEntry(p,start,TMath::Nint(TMath::Log(end)/TMath::Log(10)), 0, TGNumberFormat::kNESInteger, TGNumberFormat::kNEAAnyNumber, TGNumberFormat::kNELLimitMinMax, start, end);
        }
        fNumber->SetNumber(position);
        fNumber->Connect("ValueSet(Long_t)", "TextedHSlider", this, "MoveSlider()");
    }
    else
    {
        fSlider = new TGHSlider(p,w,type,id,options,back);
        fSlider->SetRange(TMath::Nint(start*1000),TMath::Nint(end*1000));
        fSlider->SetPosition(TMath::Nint(position*1000));
        fSlider->Connect("PositionChanged(Int_t)", "TextedHSlider", this, "MoveRealThreeNumber(Int_t)");

        fNumber = new TGNumberEntry(p,start,5, 0, TGNumberFormat::kNESRealThree, TGNumberFormat::kNEAAnyNumber, TGNumberFormat::kNELLimitMinMax, start, end);
        fNumber->SetNumber(position);
        fNumber->Connect("ValueSet(Long_t)", "TextedHSlider", this, "MoveRealThreeSlider()");
    }
}


TextedHSlider::~TextedHSlider()
{

}

TGHSlider* TextedHSlider::GetSlider()
{
    return fSlider;
}

TGNumberEntry* TextedHSlider::GetNumberEntry()
{
    return fNumber;
}

void TextedHSlider::MoveSlider()
{
    fSlider->SetPosition(TMath::Nint(fNumber->GetNumber()));
    fSlider->PositionChanged(TMath::Nint(fNumber->GetNumber()));
}

void TextedHSlider::MoveRealThreeSlider()
{
    fSlider->SetPosition(TMath::Nint(fNumber->GetNumber()*1000));
    fSlider->PositionChanged(TMath::Nint(fNumber->GetNumber()*1000));
}

void TextedHSlider::MoveNumber(Int_t pos)
{
    fNumber->SetIntNumber(pos);
    //DO not emit signal, otherwise infinite loop
}

void TextedHSlider::MoveRealThreeNumber(Int_t pos)
{
    fNumber->SetNumber(pos/1000.);
    //DO not emit signal, otherwise infinite loop
}
