/**
 * @file
 * SFMC01+SIS1160 threaded controller class implementation
 */

#pragma GCC diagnostic ignored "-Wold-style-cast"
#pragma GCC diagnostic ignored "-Wzero-as-null-pointer-constant"
#pragma GCC diagnostic ignored "-Weffc++"
#pragma GCC diagnostic ignored "-Wsign-conversion"
#pragma GCC diagnostic ignored "-Wconversion"
#pragma GCC diagnostic ignored "-Wcast-qual"
#include "ThSFMC01.hpp"
#include "TThread.h"
#include "TStopwatch.h"
#pragma GCC diagnostic warning "-Wold-style-cast"
#pragma GCC diagnostic warning "-Wzero-as-null-pointer-constant"
#pragma GCC diagnostic warning "-Weffc++"
#pragma GCC diagnostic warning "-Wsign-conversion"
#pragma GCC diagnostic warning "-Wconversion"
#pragma GCC diagnostic warning "-Wcast-qual"
#include <chrono>

uint32_t ThSFMC01::kDMASize[2] = {0, 0};
uint64_t ThSFMC01::kRAMSize[2] = {0, 0};
uint32_t* ThSFMC01::kBufferRAM[2] = {nullptr, nullptr};

ThSFMC01::ThSFMC01(const uint16_t deviceNr) :
RThread("ThSFMC01"),
fQObject()
{

}

ThSFMC01::~ThSFMC01()
{

}

void ThSFMC01::ChunkRead(ULong64_t ch, ULong64_t cNum, ULong64_t writePos, ULong64_t dFill, ULong64_t readTime, ULong64_t startTime)
{
    EmitVA("ChunkRead(ULong64_t, ULong64_t, ULong64_t, ULong64_t, ULong64_t, ULong64_t)", 6, ch, cNum, writePos, dFill, readTime, startTime);
}

void ThSFMC01::Finished()
{
    Emit("Finished()");
}

void ThSFMC01::ThreadFunction()
{
    TThread::Printf("ThSFMC01 started.");

    const uint32_t dmaMaxSize = 0x40000000; // in bytes, for each channel. DMA has a a maximum 1 GByte per channel
    const uint32_t dmaSize[2] = {kDMASize[0], kDMASize[1]};
    const uint64_t ramSize[2] = {kRAMSize[0], kRAMSize[1]};
    const uint32_t startBlock[2] = {0, 0x1000000};
    uint32_t* const bufferRAM[2] = {kBufferRAM[0], kBufferRAM[1]};
    const bool enabled[2] = {ramSize[0]>0, ramSize[1]>0};
    if(!enabled[0] && !enabled[1])
    {
        return;
    }
    for(uint16_t ch = 0; ch <= 1; ++ch)
    {
        //std::cout << dmaSize[ch] << " " << ramSize[ch] << " " << bufferRAM[ch] << std::endl;
        if(!enabled[ch])
        {
            continue;
        }
        if(ramSize[ch]%dmaSize[ch] != 0)
        {
            TThread::Printf("Error not aligned RAM buffer on channel %u, 0x%luX // 0x%08X.", ch, ramSize[ch], dmaSize[ch]);
            return;
        }
        if(dmaMaxSize%dmaSize[ch] != 0)
        {
            TThread::Printf("Error not aligned DMA buffer on channel %u, 0x%08X // 0x%08X.", ch, dmaSize[ch], dmaMaxSize);
            return;
        }
        if(dmaSize[ch]== 0)
        {
            TThread::Printf("Error zero DMA size on channel %u", ch);
            return;
        }
    }
    const uint16_t maxChunks[2] = {static_cast<uint16_t>(enabled[0] ? ramSize[0]/dmaSize[0] : 0), static_cast<uint16_t>(enabled[1] ? ramSize[1]/dmaSize[1] : 0)};

    static uint16_t chunk[2];
    if(!fPauseFlag)
    {
        chunk[0] = 0;
        chunk[1] = 0;
    }
    const uint16_t maxPeriod[2] = {static_cast<uint16_t>(enabled[0] ? dmaMaxSize/dmaSize[0] : 0), static_cast<uint16_t>(enabled[1] ? dmaMaxSize/dmaSize[1] : 0)};
    uint32_t deltaAddr[2] = {0, 0};
    uint32_t delta[2] = {0, 0};
    uint32_t nextFill[2] = {0, 0};
    uint64_t writeIndex[2] = {0, 0};
    bool readoutDone[2] = {false, false};
    TStopwatch readTime[2];
    fPauseFlag = false;
    bool overflow = false;

    const std::chrono::steady_clock::time_point tbegin = std::chrono::steady_clock::now();
    do
    {
        usleep(1000000);
        fStopFlag = true;
        /*
        for(uint16_t ch = 0; ch <= 1; ++ch)
        {
            if(!enabled[ch])
            {
                continue;
            }
            if(chunk[ch] >= maxChunks[ch])//overflow
            {
                fStopFlag = true;
                overflow = true;
                continue;
            }
            delta[ch] = static_cast<uint32_t>((chunk[ch]+1)%maxPeriod[ch]);
            deltaAddr[ch] = dmaSize[ch]/32/2*delta[ch];
            //fMutex.Lock();
            const uint32_t cba = GetCurrentBlockAddress(ch);
            //TThread::Printf("Ch %u. Chunk: %u / %u, Period: %u / %u, CurrBlock: 0x%08X\t CurrThresh: 0x%0X", ch, chunk[ch], maxChunks[ch], chunk[ch]%maxPeriod[ch], maxPeriod[ch], cba, startBlock[ch]+deltaAddr[ch]);
            if(deltaAddr[ch]>0)
            {
                if(cba>=startBlock[ch]+deltaAddr[ch])
                {
                    readTime[ch].Start();
                    int ret = 0;//ReadMemory(startBlock[ch]*32*2+(chunk[ch]%maxPeriod[ch])*dmaSize[ch],dmaSize[ch],&bufferRAM[ch][writeIndex[ch]]);
                    readTime[ch].Stop();
                    //TThread::Printf("Ch %u. Chunk: %u / %u, Period: %u / %u, CurrBlock: 0x%08X\t CurrThresh: 0x%08X", ch, chunk[ch], maxChunks[ch], chunk[ch]%maxPeriod[ch], maxPeriod[ch], cba, startBlock[ch]+deltaAddr[ch]);
                    if(ret != 0)
                    {
                        TThread::Printf("sis1160_readMemory failed with status code %i", ret);
                    }
                    readoutDone[ch] = true;
                    nextFill[ch] = cba - startBlock[ch] - deltaAddr[ch];
                    chunk[ch]++;
                    writeIndex[ch] = chunk[ch]*static_cast<uint64_t>(dmaSize[ch])/sizeof(uint32_t);
                }
                else
                {
                    readoutDone[ch] = false;
                }
            }
            else//Wrap around
            {
                if(cba>=startBlock[ch] && cba<startBlock[ch]+static_cast<uint32_t>(maxPeriod[ch]-1)*dmaSize[ch]/32/2)
                {
                    readTime[ch].Start();
                    int ret = 0;//ReadMemory(startBlock[ch]*32*2,dmaSize[ch],&bufferRAM[ch][writeIndex[ch]]);
                    readTime[ch].Stop();
                    //TThread::Printf("Ch %u. Chunk: %u / %u, Period: %u / %u, CurrBlock: 0x%08X\t CurrThresh: 0x%08X", ch, chunk[ch], maxChunks[ch], chunk[ch]%maxPeriod[ch], maxPeriod[ch], cba, startBlock[ch]+deltaAddr[ch]);
                    if(ret != 0)
                    {
                        TThread::Printf("sis1160_readMemory failed with status code %i", ret);
                    }
                    readoutDone[ch] = true;
                    nextFill[ch] = cba - startBlock[ch];
                    chunk[ch]++;
                    writeIndex[ch] = static_cast<uint64_t>(dmaSize[ch])/sizeof(uint32_t) * chunk[ch];
                }
                else
                {
                    readoutDone[ch] = false;
                }
            }
            //fMutex.UnLock();

            if(readoutDone[ch])
            {
				const std::chrono::steady_clock::time_point tend = std::chrono::steady_clock::now();
                ChunkRead(ch, chunk[ch], static_cast<uint64_t>(chunk[ch])*dmaSize[ch], nextFill[ch]*32*2, static_cast<ULong64_t>(readTime[ch].RealTime()*1000000), static_cast<unsigned long>(std::chrono::duration_cast<std::chrono::milliseconds>(tend - tbegin).count()));
            }
        }

        if(!readoutDone[0] && !readoutDone[1] && !fStopFlag && !fPauseFlag)
        {
            TThread::Sleep(0,1000);
        }*/


        if(fStopFlag)
        {
            if(overflow)
            {
                TThread::Printf("ThSFMC01 max chunk reached, overflow.");
            }
            else
            {
                //to-do: last readout read remaining data of each channel
            }
        }
    }
    while(!fStopFlag && !fPauseFlag);
    //TurnUserLED(false);

    if(fStopFlag)
    {
        TThread::Printf("ThSFMC01 received stop flag.");
    }
    if(fPauseFlag)
    {
        TThread::Printf("ThSFMC01 received pause flag.");
    }
    Finished();
    return;
}
