#pragma GCC diagnostic ignored "-Wold-style-cast"
#pragma GCC diagnostic ignored "-Wzero-as-null-pointer-constant"
#pragma GCC diagnostic ignored "-Weffc++"
#pragma GCC diagnostic ignored "-Wsign-conversion"
#pragma GCC diagnostic ignored "-Wconversion"
#pragma GCC diagnostic ignored "-Wcast-qual"
#include "TApplication.h"
#include "TROOT.h"
#include "TSystem.h"
#include "MainWindow.hpp"
#include "PulseSurfer.hpp"
#pragma GCC diagnostic warning "-Wold-style-cast"
#pragma GCC diagnostic warning "-Wzero-as-null-pointer-constant"
#pragma GCC diagnostic warning "-Weffc++"
#pragma GCC diagnostic warning "-Wsign-conversion"
#pragma GCC diagnostic warning "-Wconversion"
#pragma GCC diagnostic warning "-Wcast-qual"
#include <iostream>
#include "CLI/App.hpp"
#include "CLI/Formatter.hpp"
#include "CLI/Config.hpp"


int main(int argc, char** argv)
{
    ROOT::EnableThreadSafety();// https://root-forum.cern.ch/t/error-in-treentrantrwlock-writeunlock-write-lock-already-released-tthread-member-function/43663/6

    CLI::App app{"Run zdt-daq GUI"};

    int prbs{0};
    app.add_option("-p,--prbs", prbs, "Perform PRBS test protocol")->check(CLI::Range(3,32));
    int tpat{-1};
    app.add_option("-t,--testpattern", tpat, "Perform ADC test pattern")->check(CLI::Range(0,15));
    bool dptc{false};
    app.add_flag("--dptc", dptc, "Compress ADC data stream using DPTC");
    bool gz{false};
    app.add_flag("--gz", gz, "Save to HDD as compressed gz file");
    TString outFolder{"/tmp"};
    app.add_option("-o,--outfolder", outFolder, "Output directory where to store data")->check(CLI::ExistingDirectory);
    bool quit{false};
    app.add_flag("-q,--quit", quit, "Save and quit automatically after DAQ stops");
    CLI11_PARSE(app, argc, argv);
    if(prbs!=0 && tpat>=0)
    {
        std::cout << "Not possible to run at the same time PRBS and ADC tests" << std::endl;
        return -1;
    }

    // Verify process duplicity
    const TObjArray ps = *(gSystem->GetFromPipe("pidof runGUI").Tokenize(" "));
    if(ps.GetEntries()==0)
    {
        std::cout << "Warning: no process found with this name. You might be running valgrind" << std::endl;
    }
    else
    {
        if(ps.GetEntries()>1)
        {
            std::cout << "Preexisting process found with this name" << std::endl;
            return -2;
        }
        if(std::stoi(ps.Last()->GetName())!=gSystem->GetPid())
        {
            std::cout << "Mismatch in process ID" << std::endl;
            return -1;
        }

        std::cout << "Starting GUI with process ID " << gSystem->GetPid() << std::endl;
    }

    TApplication theApp("App", &argc, argv);

    if (gROOT->IsBatch()) {
       fprintf(stderr, "%s: cannot run in batch mode\n", argv[0]);
       return 1;
    }

    PulseSurfer surfer;

    uint32_t vendor = 0;
    uint32_t id = 0;
    MainWindow window(gClient->GetRoot(),800,600, &surfer, vendor, id, quit);
    window.SetOutputFolder(outFolder);

    if(prbs)
    {
        const uint64_t maxBurstCycles = 100000000;

        if(pow(2,prbs) > maxBurstCycles)
        {
            std::cerr << "using infinity PN" << prbs << " as max burst cycle is below period" << std::endl;
        }
        window.SetupPRBS("PN"+std::to_string(prbs), dptc, gz);
        window.AcquirePRBS();
    }
    else if(tpat >= 0)
    {
        if(tpat >8 && tpat <15)
        {
            std::cerr << "skipping pattern " << tpat << " as not yet used" << std::endl;
            if(quit)
            {
                return 0;
            }
        }
        else
        {
            window.SetupADCTest(tpat, dptc, gz);
            window.AcquireADCTest();
        }
    }
    theApp.Run();

    return 0;
}
